import numpy as np
from scipy import special as spspecial
from multiprocessing import Pool, sharedctypes, Value
import copy
from ..utils import choice
from ..utils import check_input

from sklearn.base import BaseEstimator, RegressorMixin, ClassifierMixin, ClusterMixin, TransformerMixin


def act_funcs_from_string(input_act_funcs, n):
    """
    translate activation functions input from user
    Inputs:
    input_act_funcs - list of activation functions or just one function. should be same
    length as nneurons if a list; allowed functions as of Jan 2015 are
          * tanh
          * logistic
    n - number of activation functions
    Outputs:
    list of activation functions to be used and their gradients

    """
    act_func_dict = {'tanh':[np.tanh,lambda x: np.power(np.cosh(x),-2)],
        'logistic':[spspecial.expit,lambda x: spspecial.expit(x) * (1- spspecial.expit(x))]}
    if isinstance(input_act_funcs,list):
        if len(input_act_funcs) != n:
            raise Exception('List of activation function is of different \
                length from list of neuron numbers')
        else:
            return map(lambda x: act_func_dict[x],input_act_funcs)
    else:
        return [act_func_dict[input_act_funcs]]*n

def np_to_c(np_array):
    """
    Converts numpy array to c-type to allow for sharing across different threads
    :param np_array:
    :return: ctype array
    """
    tmp = np.ctypeslib.as_ctypes(np_array)
    c_array = sharedctypes.Array(tmp._type_, tmp, lock=False)
    return c_array

def initialize_weights(complete_nneurons,n_hist):
    """
    Initialize weights randomly using numpy's random number generator
    Inputs:
    complete_nneurons - list of number of neurons (including input and output layers)
    Output:
    list of initialized weights
    """
    weights = []
    for prev_nneurons,nneurons in zip(complete_nneurons[:-1],complete_nneurons[1:]):
        weight = np_to_c(np.random.rand(prev_nneurons,nneurons) - 0.5)
        bias = np_to_c(np.random.rand(nneurons,1) - 0.5)
        weights.append([weight,bias])

    nphist = np.empty((n_hist,complete_nneurons[-1]+1))
    nphist[:,:-1] = np.inf
    nphist[:,-1] = 0
    return weights,np_to_c(nphist)

def initialize_data(np_X_train,np_Y_train,np_X_test,np_Y_test):
    """
    Initializes training and testing data
    :param np_X_train: numpy array
    :param np_Y_train: numpy array
    :param np_X_test: numpy array
    :param np_Y_test: numpy array
    :return: tuple of c-type arrays for features_training, targets_training, features_testing, targets_testing
    """
    c_X_train = np_to_c(np_X_train)
    c_Y_train = np_to_c(np_Y_train)
    c_X_test = np_to_c(np_X_test)
    c_Y_test = np_to_c(np_Y_test)
    return c_X_train,c_Y_train,c_X_test,c_Y_test

def feed_forward(X,weights,act_funcs):
    """
    Feed forward through neural network
    Inputs:
    X - input
    weights - weights
    act_funcs - activation functions
    Output:
    his - output from each layer of the network
    dhis - h'(a) for each layer in the network, where h' is the derivative of the activation
    function, and a is the sum of weighted outputs from the previous layer for use in back propagation

    """

    his = [X]  # outputs of neurons at all layers, initialized to input layer
    dhis = [np.ones((X.shape[0],X.shape[1]))]
    hi = his[0]  # output of input layer, de-reffed for efficiency
    for i, [weight, act_func] in enumerate(zip(weights[:-1], act_funcs)):
        arg = np.dot(hi, weight[0]) + weight[1].T  # input to activation function
        hi = act_func[0](arg)  # output at layer i of network
        dhi = act_func[1](arg)  # derivative of hi wrt arg
        his.append(hi)
        dhis.append(dhi)
    his.append(np.dot(his[-1], weights[-1][0]) + weights[-1][1].T)
    return his, dhis

def _output(X, weights, act_funcs):
    """
    Internal function. Gives the output of the network without book keeping each layer
    :param X: Inputs
    :param weights: type list of lists : weights
    :param act_funcs: type list : activation functions
    :return: type array : output of network

    """

    hi = X
    for weight,act_func in zip(weights[:-1],act_funcs):
        hi = act_func[0](np.dot(hi,weight[0]) + weight[1].T)
    return np.dot(hi, weights[-1][0]) + weights[-1][1].T

def output(X,nn):
    """
    User accessible output for neural network given trained weights.
    :param X: type array:Input features
    :param nn: type dict : Neural network: keys required weights and activation functions
    :return: type array: output values
    """
    return _output(X,nn['weights'],act_funcs_from_string(nn['act_funcs'],len(nn['weights'])-1))

def _c_to_numpy_weights(cweights):
    """
    Convenience function to convert shared ctype weights to numpy weights
    :param cweights:
    :return: type list of lists: lists of weights and biases of each layer.
    :note: This still shares memory with ctype weights
    """
    weights = []
    for cweight,cbias in cweights:
        weight = np.ctypeslib.as_array(cweight)
        bias = np.ctypeslib.as_array(cbias)
        weights.append([weight,bias])
    return weights

def back_prop_init(c_X_train,c_Y_train,c_X_test,c_Y_test,shared_weights,shared_hist,input_act_funcs):
    """
    Initializing back propagation.  Initializes global vars for each thread, and makes sure that all the
    threads share memory to save memory.
    :param c_X_train: ctype array of the training features
    :param c_Y_train: ctype array of the training targets
    :param c_X_test: ctype array of the testing features
    :param c_Y_test: ctype array of the testing targets
    :param shared_weights: list of weights
    :param shared_hist: array : history for early stopping
    :param input_act_funcs: user input of activation functions
    :return:
    """
    # TODO: sanitize activation function input

    global weights
    weights = _c_to_numpy_weights(shared_weights)
    global act_funcs
    act_funcs = act_funcs_from_string(input_act_funcs,len(weights)-1)
    global hist
    hist = np.ctypeslib.as_array(shared_hist)
    global X_train
    X_train = np.ctypeslib.as_array(c_X_train)
    global Y_train
    Y_train = np.ctypeslib.as_array(c_Y_train)
    global X_test
    X_test = np.ctypeslib.as_array(c_X_test)
    global Y_test
    Y_test = np.ctypeslib.as_array(c_Y_test)

def initialize_rms_weights(weights):
    """
    Initializes weights for rms prop
    :param weights: type list of lists: weights in each layer
    :return: zeroed list of lists with the same dimensions as the input list of lists
    """
    rms_weights = copy.deepcopy(weights)
    for weight,bias in rms_weights:
        weight[:] = 0
        bias[:] = 0
    return rms_weights

def prevent_underflow(np_array):
    """
    Stops underflow by replacing 0 with smallest numpy value
    :param np_array:
    :return:
    """
    zero_inds = (np_array == 0)
    if zero_inds.any():
        np_array[zero_inds] = float(np.finfo(np.float64).tiny)
    return np_array

def back_propagate(para_input):
    """
    Performs back propagation on one thread to train the neural network
    :param para_input: type list: input parameters (unpacking needed due to python's handling)
    """
    pool_index, train_start, train_end, learn_rate, rms_decay, n_epochs, \
        batch_size, n_cores, n_hist, n_check, threshold, print_level = para_input
    hist_count = 0
    rms_weights = initialize_rms_weights(weights)

    for i in range(n_epochs):
        if i % n_check == 0:
            cur_err = np.mean(np.power(_output(X_test, weights, act_funcs) - Y_test, 2), axis=0)
            if print_level > 0:
                print i, cur_err

            if np.mean(hist[:, -1]) >= threshold:
                break
            else:
                hist_index = hist_count * n_cores + pool_index
                hist[hist_index, :-1] = cur_err
                if np.mean(np.median(hist, axis=0)[:-1] - cur_err) <= 0:
                    hist[hist_index, -1] = 1
                else:
                    hist[hist_index, -1] = 0
                hist_count = (hist_count + 1) % n_hist

        for start_ind in range(train_start, train_end-1, batch_size):
            X_train_batch = X_train[start_ind:start_ind+batch_size]
            Y_train_batch = Y_train[start_ind:start_ind+batch_size]

            his,dhis = feed_forward(X_train_batch,weights,act_funcs)
            di = his[-1] - Y_train_batch

            for weight,hi,dhi,rms_weight in zip(weights[::-1],his[:-1][::-1],dhis[::-1],rms_weights[::-1]):
                weight_grad = np.dot(hi.T,di)/di.shape[0]
                bias_grad = np.mean(di,axis=0)[:,None]
                rms_weight[0] = prevent_underflow((1 - rms_decay) * np.power(weight_grad,2) + rms_decay * rms_weight[0])
                rms_weight[1] = prevent_underflow((1 - rms_decay) * np.power(bias_grad,2) + rms_decay * rms_weight[1])
                weight[0] -= learn_rate * weight_grad / np.power(rms_weight[0],0.5)
                weight[1] -= learn_rate * bias_grad / np.power(rms_weight[1],0.5)
                if np.isnan(weight[0]).any() or np.isinf(weight[0]).any() or np.isnan(weight[1]).any() \
                    or np.isinf(weight[1]).any():
                    break
                di = dhi * np.dot(di,weight[0].T)

# @profile
def train(X,Y,nneurons,input_act_funcs,validation_size=0.2,learn_rate=0.001,rms_decay=0.9,n_epochs=10000,
    batch_size=256,n_cores=1,n_hist=20,n_check=50,threshold=0.1, print_level=1):
    """
    Main training function
    :param X: pandas dataframe or numpy array
        input training data
    :param Y: pandas dataframe or numpy array
        output training data
    :param nneurons: list of integers
        describing how many neurons there are in each layer
    :param input_act_funcs: list of activation functions or just one function (string)
        should be same length as nneurons if a list
    :param validation_size: float between zero and one, optional (default = 0.2)
        size of data to be selected randomly for validation

    :return nnet: a tuple with trained weights and the activation functions

    """
    X_train, _ = check_input(X,'Training input', format_out='ar')
    Y_train, _ = check_input(Y,'Training output',n0=X_train.shape[0],format_out='ar')
    X_train, X_test, Y_train, Y_test = choice(X_train, Y_train, n = validation_size)
    X_test, _ = check_input(X_test,'Testing input',n1=X_train.shape[1],format_out='ar')
    Y_test, _ = check_input(Y_test,'Testing output',n0=X_test.shape[0],n1=Y_train.shape[1],format_out='ar')
    n_features = X_train.shape[1]
    n_outputs = Y_train.shape[1]
    N = X_train.shape[0]

    c_X_train, c_Y_train, c_X_test, c_Y_test = initialize_data(X_train,Y_train,X_test,Y_test)
    del X_train
    del X_test
    del Y_train
    del Y_test

    cweights, chist = initialize_weights([n_features] + nneurons +
        [n_outputs],n_hist*n_cores)

    big_batch_size = (N / n_cores) + 1
    para_inputs = [[i, j, j+big_batch_size,learn_rate,rms_decay,n_epochs,batch_size,n_cores,n_hist,
        n_check,threshold, print_level] for i,j in enumerate(range(0,N,big_batch_size))]
    pool = Pool(processes=n_cores, initializer=back_prop_init, initargs=(c_X_train,c_Y_train,c_X_test,c_Y_test,
        cweights, chist, input_act_funcs))
    pool.map(back_propagate, para_inputs)

    # use this to test individual thread
    # back_prop_init(c_X_train,c_Y_train,c_X_test,c_Y_test,cweights, chist, input_act_funcs)
    # back_propagate([0,0,N,learn_rate,rms_decay,n_epochs,batch_size,1,n_hist,n_check,threshold])

    final_weights = _c_to_numpy_weights(cweights)
    return {'weights':final_weights,'act_funcs':input_act_funcs}

class mlp_hogwild(BaseEstimator, RegressorMixin):
    def __init__(self,nneurons,input_act_funcs,validation_size=0.2,learn_rate=0.001,rms_decay=0.9,n_epochs=10000,
    batch_size=256,n_cores=1,n_hist=20,n_check=50,threshold=0.1, print_level=1):
        self.nneurons = nneurons
        self.input_act_funcs = input_act_funcs
        self.validation_size = validation_size
        self.learn_rate = learn_rate
        self.rms_decay = rms_decay
        self.n_epochs = n_epochs
        self.batch_size = batch_size
        self.n_cores = n_cores
        self.n_hist = n_hist
        self.n_check = n_check
        self.threshold = threshold
        self.print_level = print_level
        # model will be defined only after fitting
        self.model = None

    def fit(self,X,y):
        self.model = train(X,y,self.nneurons, self.input_act_funcs, self.validation_size, self.learn_rate, self.rms_decay,
                           self.n_epochs, self.batch_size, self.n_cores, self.n_hist, self.n_check, self.threshold, self.print_level)
    def predict(self, X):
        if self.model is None:
            msg = "This mlp_hogwild instance is not fitted yet. Call 'fit' with appropriate arguments before using this method."
            raise RuntimeError(msg)
        return output(X,self.model)
    def score(self, X, y):
        y_pred = self.predict(X)
        return np.mean(np.abs(y - y_pred))


# if __name__ == '__main__':
#     # import matplotlib.pyplot as plt
#     N = 2000
#     X = (np.random.rand(N,1) - 0.5) * 10
#     Y = np.sin(X)
#     mlp = mlp_hogwild([20,10],'tanh',validation_size=0.1,learn_rate=0.001,n_epochs=50000,n_cores=2)
#     mlp.fit(X,Y)
#     print mlp.score(X[:N/2],Y[:N/2])
#     # plt.plot(X_test[:,0],output(X_test,nnet)[:,0],'x')
#     # plt.plot(X_test[:,0],Y_test[:,0],'o')
#     # plt.show()



import pandas as pd
import numpy as np
import os
import warnings

from ..base import BASE

##################################################################### 2 Prepare Data

# Enter

class XYZreader(BASE):
    def fit(self):
        try:
            from cheml.initialization import XYZreader
            model = XYZreader(**self.parameters)
            molecules = model.read()
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'molecules':
                self.set_value(token, molecules)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            # elif token == 'max_n_atoms':
            #     self.set_value(token, model.max_n_atoms)
            #     self.outputs[token].count = order.count(token)
            #     self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class load_cep_homo(BASE):
    def fit(self):
        try:
            from cheml.datasets import load_cep_homo
            smiles,homo = load_cep_homo()
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'smiles':
                self.set_value(token, smiles)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'homo':
                self.set_value(token, homo)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
        del self.inputs

class load_organic_density(BASE):
    def fit(self):
        try:
            from cheml.datasets import load_organic_density
            smiles,density,features = load_organic_density()
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'smiles':
                self.set_value(token, smiles)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'density':
                self.set_value(token, density)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'features':
                self.set_value(token, features)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
        del self.inputs

class load_xyz_polarizability(BASE):
    def fit(self):
        try:
            from cheml.datasets import load_xyz_polarizability
            coordinates, pol = load_xyz_polarizability()
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'coordinates':
                self.set_value(token, coordinates)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'polarizability':
                self.set_value(token, pol)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
        del self.inputs

class load_comp_energy(BASE):
    def fit(self):
        try:
            from cheml.datasets import load_comp_energy
            entries, df = load_comp_energy()
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'entries':
                self.set_value(token, entries)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'formation_energy':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
        del self.inputs

class load_crystal_structures(BASE):
    def fit(self):
        try:
            from cheml.datasets import load_crystal_structures
            entries = load_crystal_structures()
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'entries':
                self.set_value(token, entries)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs




class ConvertFile(BASE):
    def fit(self):
        self.paramFROMinput()
        # self.required('file_path',req=True)
        # file_path=self.inputs['file_path'].value
        print 'from:', self.parameters['from_format']
        print 'to:', self.parameters['to_format']
        # if 'file_path' not in self.parameters and '@' not in file_path:
            # self.parameters['file_path']=file_path
        try:
            from cheml.initialization import ConvertFile
            model = ConvertFile(**self.parameters)
            converted_file_paths=model.convert()
        except Exception as err:
            msg='@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__+': ' +err.message
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg="@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1,self.Task,token)
                raise NameError(msg)
            elif token == 'converted_file_paths':
                self.set_value(token,converted_file_paths)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock,token)] = self.outputs[token]
        del self.inputs

class scatter2D(BASE):
    def fit(self):
        # self.paramFROMinput()
        self.required('dfx', req=True)
        dfx = self.inputs['dfx'].value
        self.required('dfy', req=True)
        dfy = self.inputs['dfy'].value
        if 'x' in self.parameters:
            x = self.parameters.pop('x')
        else:
            msg = "@Task #%i(%s): the x header or position is required" % (self.iblock + 1, self.Task)
            raise NameError(msg)
        if 'y' in self.parameters:
            y = self.parameters.pop('y')
        else:
            msg = "@Task #%i(%s): the y header or position is required" % (self.iblock + 1, self.Task)
            raise NameError(msg)
        try:
            from cheml.visualization import scatter2D
            sc = scatter2D(**self.parameters)
            fig = sc.plot(dfx,dfy,x,y)
        except Exception as err:
            msg='@Task #%i(%s): ' % (self.iblock + 1,self.Task)+ type(err).__name__+': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'fig':
                self.set_value(token, fig)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
        del self.inputs

class hist(BASE):
    def fit(self):
        # self.paramFROMinput()
        self.required('dfx', req=True)
        dfx = self.inputs['dfx'].value
        if 'x' in self.parameters:
            x = self.parameters.pop('x')
        else:
            msg = "@Task #%i(%s): the x header or position is required" % (self.iblock + 1, self.Task)
            raise NameError(msg)
        try:
            from cheml.visualization import hist
            hg = hist(**self.parameters)
            fig = hg.plot(dfx, x)
        except Exception as err:
            msg='@Task #%i(%s): ' % (self.iblock + 1,self.Task)+ type(err).__name__+': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'fig':
                self.set_value(token, fig)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
        del self.inputs

class decorator(BASE):
    def fit(self):
        # self.paramFROMinput()
        self.required('fig', req=True)
        fig = self.inputs['fig'].value
        font_params = {}
        for p in ['family','size','weight','style','variant']:
            if p in self.parameters:
                font_params[p] = self.parameters.pop(p)
        try:
            from cheml.visualization import decorator
            dec = decorator(**self.parameters)
            dec.matplotlib_font(**font_params)
            fig = dec.fit(fig)
        except Exception as err:
            msg='@Task #%i(%s): ' % (self.iblock + 1,self.Task)+ type(err).__name__+': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'fig':
                self.set_value(token, fig)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
        del self.inputs

class SavePlot(BASE):
    def fit(self):
        self.required('fig', req=True)
        fig = self.inputs['fig'].value
        # self.paramFROMinput()
        try:
            from cheml.visualization import SavePlot
            sp = SavePlot(**self.parameters)
            sp.save(fig, self.Base.output_directory)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)

        del self.inputs

class PyScript(BASE):
    def fit(self):
        # step1: check inputs
        inputs = [token for token in self.inputs if self.inputs[token].value is not None]
        for token in inputs:
            code = compile("%s = self.inputs['%s'].value"%(token,token), "<string>", "exec")
            exec code
        for line in sorted(self.parameters.keys()):
            code = compile(self.parameters[line], "<string>", "exec")
            exec code
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token == 'ov1':
                self.set_value(token, ov1)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'ov2':
                self.set_value(token, ov2)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'ov3':
                self.set_value(token, ov3)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'ov4':
                self.set_value(token, ov4)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'ov5':
                self.set_value(token, ov5)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'ov6':
                self.set_value(token, ov6)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            else:
                msg = "@Task #%i(%s): asked to send a non valid output token '%s'" % (
                    self.iblock+1,self.Task,token)
                raise NameError(msg)
        del self.inputs


# Feature Representation

class RDKitFingerprint(BASE):
    def fit(self):
        # step1: check inputs
        # step2: assign inputs to parameters if necessary (param = @token)
        self.paramFROMinput()
        if 'molfile' in self.parameters:
            molfile = self.parameters['molfile']
            self.parameters.pop('molfile')
        else:
            self.required('molfile', req=True)
            molfile = self.inputs['molfile'].value
        # step3: check the dimension of input data frame
        # step4: extract parameters
        if 'path' in self.parameters:
            path = self.parameters.pop('path')
        else:
            path = None
        if 'arguments' in self.parameters:
            arguments = self.parameters.pop('arguments')
        else:
            arguments = []

        # step5: import module and make APIs
        try:
            from cheml.chem import RDKitFingerprint
            model = RDKitFingerprint(**self.parameters)
            model.MolfromFile(molfile,path,*arguments)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)

        # step6: send out
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, model.Fingerprint())
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'removed_rows':
                self.set_value(token, model.removed_rows)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        # step7: delete all inputs from memory
        del self.inputs

class Dragon(BASE):
    def fit(self):
        # step1: check inputs
        # step2: assign inputs to parameters if necessary (param = @token)
        self.paramFROMinput()
        print 'molfile:', self.parameters['molFile']

        # step3: check the dimension of input data frame
        # step4: extract  parameters
        if 'script' in self.parameters:
            script = self.parameters.pop('script')
        else:
            script = 'new'
        # if 'output_directory' in self.parameters:
        #     output_directory = self.parameters.pop('output_directory')
        # else:
        #     output_directory = ''
        output_directory = self.Base.output_directory #+ '/' + output_directory
        if 'SaveFilePath' not in self.parameters:
            self.parameters['SaveFilePath'] = "Dragon_descriptors.txt"

        # step5: import module and make APIs
        try:
            from cheml.chem import Dragon
            model = Dragon(**self.parameters)
            model.script_wizard(script, output_directory)
            model.run()
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)

        # step6: send out
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                df_path = model.data_path
                df = pd.read_csv(df_path, sep=None, engine='python')
                df = df.drop(['No.','NAME'],axis=1)
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        # step7: delete dragon descriptors and all inputs from memory
        os.remove(model.output_directory + self.parameters['SaveFilePath'])
        del self.inputs

class CoulombMatrix(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('molecules', req=True)
        molecules = self.inputs['molecules'].value
        try:
            from cheml.chem import CoulombMatrix
            model = CoulombMatrix(**self.parameters)
            df = model.represent(molecules)
        except Exception as err:
            print err.message
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class BagofBonds(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('molecules', req=True)
        molecules = self.inputs['molecules'].value
        try:
            from cheml.chem import BagofBonds
            model = BagofBonds(**self.parameters)
            df = model.represent(molecules)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            # elif token == 'headers':
            #     self.set_value(token, model.headers)
            #     self.outputs[token].count = order.count(token)
            #     self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class DistanceMatrix(BASE):
    def fit(self):
        self.paramFROMinput()
        # check inputs
        self.required('df', req=True)
        try:
            from cheml.chem import DistanceMatrix
            model = DistanceMatrix(**self.parameters)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                ar = self.inputs['df'].value
                df = pd.DataFrame(model.transform(ar))
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
        del self.inputs

class APEAttributeGenerator(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('entries', req=True)
        entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import APEAttributeGenerator
            ape = APEAttributeGenerator()
            if 'packing_threshold' in self.parameters:
                if self.parameters['packing_threshold']:
                    ape.set_packing_threshold(self.parameters['packing_threshold'])
            if 'n_nearest_to_eval' in self.parameters:
                if self.parameters['n_nearest_to_eval']:
                    ape.set_n_nearest_to_eval(self.parameters['n_nearest_to_eval'])
            if 'radius_property' in self.parameters:
                if self.parameters['radius_property']:
                    ape.set_radius_property(self.parameters['radius_property'])
            df = ape.generate_features(entries)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class ChargeDependentAttributeGenerator(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('entries', req=True)
        entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import ChargeDependentAttributeGenerator
            cd = ChargeDependentAttributeGenerator()
            df = cd.generate_features(entries)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class ElementalPropertyAttributeGenerator(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('entries', req=True)
        entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import ElementalPropertyAttributeGenerator
            if 'elemental_properties' in self.parameters:
                elemental_properties = self.parameters.pop('elemental_properties')
            ep = ElementalPropertyAttributeGenerator(**self.parameters)
            if elemental_properties:
                ep.add_elemental_properties(elemental_properties)
            df = ep.generate_features(entries)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class ElementFractionAttributeGenerator(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('entries', req=True)
        entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import ElementFractionAttributeGenerator
            ef = ElementFractionAttributeGenerator()
            df = ef.generate_features(entries)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class ElementPairPropertyAttributeGenerator(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('entries', req=True)
        entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import ElementPairPropertyAttributeGenerator
            epp = ElementPairPropertyAttributeGenerator()
            if 'elemental_pair_properties' in self.parameters:
                if self.parameters['elemental_pair_properties']:
                    epp.add_elemental_pair_properties(
                        self.parameters['elemental_pair_properties'])
            df = epp.generate_features(entries)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class GCLPAttributeGenerator(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('entries', req=True)
        entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import GCLPAttributeGenerator
            gclp = GCLPAttributeGenerator()
            if 'count_phases' in self.parameters and self.parameters[
                'count_phases']:
                gclp.set_count_phases(self.parameters[
                'count_phases'])
            if 'phases' in self.parameters and self.parameters['phases'] and \
                    'energies' in self.parameters and self.parameters[
                'energies']:
                gclp.set_phases(self.parameters['phases'], self.parameters[
                    'energies'])
            df = gclp.generate_features(entries)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class IonicCompoundProximityAttributeGenerator(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('entries', req=True)
        entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import IonicCompoundProximityAttributeGenerator
            icp = IonicCompoundProximityAttributeGenerator()
            if 'max_formula_unit' in self.parameters and self.parameters[
                'max_formula_unit'] != 14:
                icp.set_max_formula_unit(self.parameters['max_formula_unit'])
            df = icp.generate_features(entries)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class IonicityAttributeGenerator(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('entries', req=True)
        entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import IonicityAttributeGenerator
            ig = IonicityAttributeGenerator()
            df = ig.generate_features(entries)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class MeredigAttributeGenerator(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('entries', req=True)
        entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import MeredigAttributeGenerator
            ma = MeredigAttributeGenerator()
            df = ma.generate_features(entries)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class StoichiometricAttributeGenerator(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('entries', req=True)
        entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import StoichiometricAttributeGenerator
            sg = StoichiometricAttributeGenerator()
            if 'p_norms' in self.parameters:
                if self.parameters['p_norms']:
                    sg.add_p_norms(self.parameters['p_norms'])
            df = sg.generate_features(entries)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class ValenceShellAttributeGenerator(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('entries', req=True)
        entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import ValenceShellAttributeGenerator
            vs = ValenceShellAttributeGenerator()
            df = vs.generate_features(entries)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class YangOmegaAttributeGenerator(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('entries', req=True)
        entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import YangOmegaAttributeGenerator
            yo = YangOmegaAttributeGenerator()
            df = yo.generate_features(entries)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class APRDFAttributeGenerator(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('entries', req=True)
        entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import APRDFAttributeGenerator
            aprdf = APRDFAttributeGenerator()
            if 'cut_off_distance' in self.parameters:
                if self.parameters['cut_off_distance'] != 10.0:
                    aprdf.set_cut_off_distance(
                        self.parameters['cut_off_distance'])
            if 'num_points' in self.parameters:
                if self.parameters['num_points'] != 6:
                    aprdf.set_num_points(
                        self.parameters['num_points'])
            if 'smooth_parameter' in self.parameters:
                if self.parameters['smooth_parameter'] != 4.0:
                    aprdf.set_smoothing_parameter(
                        self.parameters['smooth_parameter'])
            aprdf.add_elemental_properties(self.parameters['elemental_properties'])
            df = aprdf.generate_features(entries)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class ChemicalOrderingAttributeGenerator(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('entries', req=True)
        entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import ChemicalOrderingAttributeGenerator
            co = ChemicalOrderingAttributeGenerator()
            if 'shells' in self.parameters:
                if self.parameters['shells']:
                    co.set_shells(
                        self.parameters['shells'])
            if 'weighted' in self.parameters:
                if self.parameters['weighted']:
                    co.set_weighted(
                        self.parameters['weighted'])

            df = co.generate_features(entries)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class CoordinationNumberAttributeGenerator(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('entries', req=True)
        entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import CoordinationNumberAttributeGenerator
            cn = CoordinationNumberAttributeGenerator()
            df = cn.generate_features(entries)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class CoulombMatrixAttributeGenerator(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('entries', req=True)
        entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import CoulombMatrixAttributeGenerator
            cm = CoulombMatrixAttributeGenerator()
            if 'n_eigenvalues' in self.parameters:
                if self.parameters['n_eigenvalues']:
                    cm.set_n_eigenvalues(
                        self.parameters['n_eigenvalues'])

            df = cm.generate_features(entries)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class EffectiveCoordinationNumberAttributeGenerator(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('entries', req=True)
        entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import EffectiveCoordinationNumberAttributeGenerator
            ecn = EffectiveCoordinationNumberAttributeGenerator()
            df = ecn.generate_features(entries)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class LatticeSimilarityAttributeGenerator(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('entries', req=True)
        entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import LatticeSimilarityAttributeGenerator
            ls = LatticeSimilarityAttributeGenerator()
            df = ls.generate_features(entries)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class LocalPropertyDifferenceAttributeGenerator(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('entries', req=True)
        entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import LocalPropertyDifferenceAttributeGenerator
            lpd = LocalPropertyDifferenceAttributeGenerator()
            if 'shells' in self.parameters:
                if not (len(self.parameters['shells']) == 1 and
                            self.parameters['shells'][0] == 1):
                    lpd.add_shells(self.parameters['shells'])
            lpd.add_elemental_properties(
                self.parameters['elemental_properties'])
            df = lpd.generate_features(entries)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class LocalPropertyVarianceAttributeGenerator(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('entries', req=True)
        entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import LocalPropertyVarianceAttributeGenerator
            lpv = LocalPropertyVarianceAttributeGenerator()
            if 'shells' in self.parameters:
                if not (len(self.parameters['shells']) == 1 and
                                self.parameters['shells'][0] == 1):
                    lpv.add_shells(self.parameters['shells'])
            lpv.add_elemental_properties(
                self.parameters['elemental_properties'])
            df = lpv.generate_features(entries)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class PackingEfficiencyAttributeGenerator(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('entries', req=True)
        entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import PackingEfficiencyAttributeGenerator
            pe = PackingEfficiencyAttributeGenerator()
            df = pe.generate_features(entries)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class PRDFAttributeGenerator(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('entries', req=True)
        entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import PRDFAttributeGenerator
            prdf = PRDFAttributeGenerator()
            if 'cut_off_distance' in self.parameters:
                if self.parameters['cut_off_distance'] != 10.0:
                    prdf.set_cut_off_distance(
                        self.parameters['cut_off_distance'])
            if 'num_points' in self.parameters:
                if self.parameters['num_points'] != 20:
                    prdf.set_num_points(
                        self.parameters['num_points'])
            prdf.set_elements(entries)
            df = prdf.generate_features(entries)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class StructuralHeterogeneityAttributeGenerator(BASE):
    def fit(self):
        self.paramFROMinput()
        self.required('entries', req=True)
        entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import StructuralHeterogeneityAttributeGenerator
            sh = StructuralHeterogeneityAttributeGenerator()
            df = sh.generate_features(entries)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class CompositionEntry(BASE):
    def fit(self):
        # self.paramFROMinput()
        # self.required('entries', req=True)
        # entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import CompositionEntry
            composition_list = CompositionEntry.import_composition_list(**self.parameters)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'entries':
                self.set_value(token, composition_list)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs

class CrystalStructureEntry(BASE):
    def fit(self):
        # self.paramFROMinput()
        # self.required('entries', req=True)
        # entries = self.inputs['entries'].value
        try:
            from cheml.chem.magpie_python import CrystalStructureEntry
            structures_list = CrystalStructureEntry.import_structures_list(**self.parameters)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)
        order = [edge[1] for edge in self.Base.graph if
                 edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (
                    self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'entries':
                self.set_value(token, structures_list)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        del self.inputs


# data manipulation

class MissingValues(BASE):
    def fit(self):
        # parameters
        self.paramFROMinput()
        if 'func_method' in self.parameters:
            method = self.parameters.pop('func_method')
        else:
            method = None
        # get df value only in case method is None, but output df is requested
        df = self.inputs['df'].value

        # process
        try:
            from cheml.preprocessing import MissingValues
            if method is None:
                model = MissingValues(**self.parameters)
            elif method == 'fit_transform':
                model = MissingValues(**self.parameters)
                self.required('df', req=True)
                df = self.inputs['df'].value
                df, _ = self.data_check('df', df, ndim=2, n0=None, n1=None, format_out='df')
                df = model.fit_transform(df)
            elif method == 'transform':
                self.required('df', req=True)
                df = self.inputs['df'].value
                df, _ = self.data_check('df', df, ndim=2, n0=None, n1=None, format_out='df')
                self.required('api', req=True)
                model = self.inputs['api'].value
                df = model.transform(df)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)

        # step6: send out
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'api':
                self.set_value(token, model)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
        # step7: delete all inputs from memory
        del self.inputs

class ConstantColumns(BASE):
    def fit(self):
        # parameters
        self.paramFROMinput()
        if 'func_method' in self.parameters:
            method = self.parameters.pop('func_method')
        else:
            method = None
        # get df value only in case method is None, but output df is requested
        df = self.inputs['df'].value

        # step4: import module and make APIs
        try:
            from cheml.preprocessing import ConstantColumns
            if method is None:
                model = ConstantColumns()
            elif method == 'fit_transform':
                model = ConstantColumns()
                self.required('df', req=True)
                df = self.inputs['df'].value
                df, _ = self.data_check('df', df, ndim=2, n0=None, n1=None, format_out='df')
                df = model.fit_transform(df)
            elif method == 'transform':
                self.required('df', req=True)
                df = self.inputs['df'].value
                df, _ = self.data_check('df', df, ndim=2, n0=None, n1=None, format_out='df')
                self.required('api', req=True)
                model = self.inputs['api'].value
                df = model.transform(df)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)

        # step5: process
        # step6: send out
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'removed_columns_':
                removed_columns_ = pd.DataFrame(model.removed_columns_)
                self.set_value(token, removed_columns_)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'api':
                self.set_value(token, model)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        # step7: delete all inputs from memory
        del self.inputs

class Outliers(BASE):
    def fit(self):
        # parameters
        self.paramFROMinput()
        method = self.parameters.pop('func_method')
        # get df value only in case method is None, but output df is requested
        df = self.inputs['df'].value

        # step4: import module and make APIs
        try:
            from cheml.preprocessing import Outliers
            if method is None:
                model = Outliers(**self.parameters)
            elif method == 'fit_transform':
                model = Outliers(**self.parameters)
                self.required('df', req=True)
                df = self.inputs['df'].value
                df, _ = self.data_check('df', df, ndim=2, n0=None, n1=None, format_out='df')
                df_out = model.fit_transform(df)
            elif method == 'transform':
                self.required('df', req=True)
                df = self.inputs['df'].value
                df, _ = self.data_check('df', df, ndim=2, n0=None, n1=None, format_out='df')
                self.required('api', req=True)
                model = self.inputs['api'].value
                df_out = model.transform(df)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)

        # step5: process
        # step6: send out
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df':
                self.set_value(token, df_out)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'removed_rows_':
                removed_columns_ = pd.DataFrame(model.removed_columns_)
                self.set_value(token, removed_columns_)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'api':
                self.set_value(token, model)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        # step7: delete all inputs from memory
        del self.inputs


# class Trimmer(BASE):
#     def legal_IO(self):
#         self.legal_inputs = {'dfx': None, 'dfy': None}
#         self.legal_outputs = {'dfx': None, 'dfy': None, 'api': None}
#         requirements = ['cheml', 'pandas']
#         self.Base.requirements += [i for i in requirements if i not in self.Base.requirements]
#
#     def fit(self):
#         from cheml.initializtion import Trimmer
#         dfx = self.type_check('dfx', cheml_type='dfx', req=True, py_type=pd.DataFrame)
#         dfy = self.type_check('dfy', cheml_type='dfy', req=True, py_type=pd.DataFrame)
#         try:
#             model = Trimmer(**self.parameters)
#             dfx, dfy = model.fit_transform(dfx,dfy)
#         except Exception as err:
#             msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
#                 err).__name__ + ': ' + str(err.message)
#             raise TypeError(msg)
#         order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
#         for token in set(order):
#             if token == 'dfx':
#                 self.Base.send[(self.iblock, token)] = [dfx, order.count(token),
#                                                         (self.iblock, token, self.Host, self.Function)]
#             elif token == 'dfy':
#                 self.Base.send[(self.iblock, token)] = [dfy, order.count(token),
#                                                         (self.iblock, token, self.Host, self.Function)]
#             elif token == 'api':
#                 self.Base.send[(self.iblock, token)] = [model, order.count(token),
#                                                         (self.iblock, token, self.Host, self.Function)]
#             else:
#                 msg = "@Task #%i(%s): asked to send a non valid output token '%s'" % (
#                     self.iblock + 1, self.Task, token)
#                 raise NameError(msg)
#         del self.legal_inputs

# class Uniformer(BASE):
#     def legal_IO(self):
#         self.legal_inputs = {'dfx': None, 'dfy': None}
#         self.legal_outputs = {'dfx': None, 'dfy': None, 'api': None}
#         requirements = ['cheml', 'pandas']
#         self.Base.requirements += [i for i in requirements if i not in self.Base.requirements]
#
#     def fit(self):
#         from cheml.initializtion import Uniformer
#         dfx = self.type_check('dfx', cheml_type='dfx', req=True, py_type=pd.DataFrame)
#         dfy = self.type_check('dfy', cheml_type='dfy', req=True, py_type=pd.DataFrame)
#         try:
#             model = Uniformer(**self.parameters)
#             dfx, dfy = model.fit_transform(dfx, dfy)
#         except Exception as err:
#             msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
#                 err).__name__ + ': ' + str(err.message)
#             raise TypeError(msg)
#         order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
#         for token in set(order):
#             if token == 'dfx':
#                 self.Base.send[(self.iblock, token)] = [dfx, order.count(token),
#                                                         (self.iblock, token, self.Host, self.Function)]
#             elif token == 'dfy':
#                 self.Base.send[(self.iblock, token)] = [dfy, order.count(token),
#                                                         (self.iblock, token, self.Host, self.Function)]
#             elif token == 'api':
#                 self.Base.send[(self.iblock, token)] = [model, order.count(token),
#                                                         (self.iblock, token, self.Host, self.Function)]
#             else:
#                 msg = "@Task #%i(%s): asked to send a non valid output token '%s'" % (
#                     self.iblock + 1, self.Task, token)
#                 raise NameError(msg)
#         del self.legal_inputs


# data manipulation

class Split(BASE):
    def fit(self):
        # step1: check inputs
        self.required('df', req=True)
        df = self.inputs['df'].value

        # step2: assign inputs to parameters if necessary (param = @token)
        self.paramFROMinput()

        # step3: check the dimension of input data frame
        df, _ = self.data_check('df', df, ndim=2, n0=None, n1=None, format_out='df')

        # step4: import module and make APIs
        try:
            from cheml.initialization import Split
            split = Split(**self.parameters)
            df1, df2 = split.fit(df)
        except Exception as err:
            msg = '@Task #%i(%s): '%(self.iblock+1, self.Task) + type(err).__name__ + ': '+ str(err.message)
            raise TypeError(msg)

        # step5: process
        # step6: send out
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'df1':
                self.set_value(token, df1)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'df2':
                self.set_value(token, df2)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        # step7: delete all inputs from memory
        del self.inputs



##################################################################### 3 Define Model

# Regression

class MLP(BASE):
    def fit(self):
        # step2: assign inputs to parameters if necessary (param = @token)
        self.paramFROMinput()
        if 'func_method' in self.parameters:
            method = self.parameters.pop('func_method')
        else:
            method = None

        # step4: import module and make APIs
        try:
            from cheml.nn.keras import MLP
            if method is None:
                model = MLP(**self.parameters)
            elif method == 'fit':
                model = MLP(**self.parameters)
                self.required('dfx', req=True)
                dfx = self.inputs['dfx'].value
                dfx, _ = self.data_check('dfx', dfx, ndim=2, n0=None, n1=None, format_out='ar')
                self.required('dfy', req=True)
                dfy = self.inputs['dfy'].value
                dfy, _ = self.data_check('dfy', dfy, ndim=2, n0=dfx.shape[0], n1=None, format_out='ar')
                model.fit(dfx,dfy)
            elif method == 'predict':
                self.required('dfx', req=True)
                self.required('api', req=True)
                dfx = self.inputs['dfx'].value
                dfx, _ = self.data_check('dfx', dfx, ndim=2, n0=None, n1=None, format_out='ar')
                model = self.inputs['api'].value
                dfy_predict = model.predict(dfx)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                    err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)

        # step5: process
        # step6: send out
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'api':
                self.set_value(token, model)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'dfy_predict':
                self.set_value(token, pd.DataFrame(dfy_predict))
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        # step7: delete all inputs from memory
        del self.inputs

class MLP_sklearn(BASE):
    def fit(self):
        # step2: assign inputs to parameters if necessary (param = @token)
        self.paramFROMinput()
        if 'func_method' in self.parameters:
            method = self.parameters.pop('func_method')
        else:
            method = None

        # step4: import module and make APIs
        try:
            from cheml.nn.keras import MLP_sklearn
            if method is None:
                model = MLP_sklearn(**self.parameters)
            elif method == 'fit':
                model = MLP_sklearn(**self.parameters)
                self.required('dfx', req=True)
                dfx = self.inputs['dfx'].value
                dfx, _ = self.data_check('dfx', dfx, ndim=2, n0=None, n1=None, format_out='ar')
                self.required('dfy', req=True)
                dfy = self.inputs['dfy'].value
                dfy, _ = self.data_check('dfy', dfy, ndim=2, n0=dfx.shape[0], n1=None, format_out='ar')
                model.fit(dfx,dfy)
            elif method == 'predict':
                self.required('dfx', req=True)
                self.required('api', req=True)
                dfx = self.inputs['dfx'].value
                dfx, _ = self.data_check('dfx', dfx, ndim=2, n0=None, n1=None, format_out='ar')
                model = self.inputs['api'].value
                dfy_predict = model.predict(dfx)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                    err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)

        # step5: process
        # step6: send out
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'api':
                self.set_value(token, model)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'dfy_predict':
                self.set_value(token, pd.DataFrame(dfy_predict))
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        # step7: delete all inputs from memory
        del self.inputs

# class mlp_hogwild(BASE):
#     def fit(self):
#         # step1: check inputs
#         self.required('dfx_train', req=True)
#         dfx_train = self.inputs['dfx_train'].value
#         self.required('dfy_train', req=True)
#         dfy_train = self.inputs['dfy_train'].value
#
#         # step2: assign inputs to parameters if necessary (param = @token)
#         self.paramFROMinput()
#         method = self.parameters.pop('func_method')
#
#         # step3: check the dimension of input data frame
#         dfx_train, _ = self.data_check('dfx_train', dfx_train, ndim=2, n0=None, n1=None, format_out='ar')
#         dfy_train, _ = self.data_check('dfy_train', dfy_train, ndim=2, n0=dfx_train.shape[0], n1=None, format_out='ar')
#
#         # step4: import module and make APIs
#         try:
#             from cheml.nn import mlp_hogwild
#             if method is None:
#                 model = mlp_hogwild(**self.parameters)
#             elif method == 'fit':
#                 model = mlp_hogwild(**self.parameters)
#                 self.required('dfx', req=True)
#                 dfx = self.inputs['dfx'].value
#                 dfx, _ = self.data_check('dfx', dfx, ndim=2, n0=None, n1=None, format_out='ar')
#                 self.required('dfy', req=True)
#                 dfy = self.inputs['dfy'].value
#                 dfy, _ = self.data_check('dfy', dfy, ndim=2, n0=dfx.shape[0], n1=None, format_out='ar')
#                 model.fit(dfx,dfy)
#             elif method == 'predict':
#                 self.required('dfx', req=True)
#                 self.required('api', req=True)
#                 dfx = self.inputs['dfx'].value
#                 dfx, _ = self.data_check('dfx', dfx, ndim=2, n0=None, n1=None, format_out='ar')
#                 api = self.inputs['api'].value
#                 dfy_predict = api.predict(dfx)
#         except Exception as err:
#             msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
#                     err).__name__ + ': ' + str(err.message)
#             raise TypeError(msg)
#
#         # step5: process
#         # step6: send out
#         order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
#         for token in set(order):
#             if token not in self.outputs:
#                 msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
#                 raise NameError(msg)
#             elif token == 'api':
#                 self.set_value(token, model)
#                 self.outputs[token].count = order.count(token)
#                 self.Base.send[(self.iblock, token)] = self.outputs[token]
#             elif token == 'dfy_predict':
#                 self.set_value(token, dfy_predict)
#                 self.outputs[token].count = order.count(token)
#                 self.Base.send[(self.iblock, token)] = self.outputs[token]
#
#         # step7: delete all inputs from memory
#         del self.inputs
#
# class mlp_dsgd(BASE):
#     # must be run with slurm script
#     # Todo: first fix the slurm script function at cheml.initialization
#     # Todo: then embede the slurm commands in this class to run the slurm script
#     # Todo: or make the slurm script in this function too
#     def legal_IO(self):
#         self.legal_inputs = {'dfx_train': None, 'dfx_test': None, 'dfy_train': None, 'dfy_test': None}
#         self.legal_outputs = {'dfy_train_pred': None, 'model': None}
#         requirements = ['cheml', 'pandas']
#         self.Base.requirements += [i for i in requirements if i not in self.Base.requirements]
#
#     def fit(self):
#         from cheml.nn import nn_dsgd
#         cheml_type = "%s_%s" % (self.Base.graph_info[self.iblock][0], self.Base.graph_info[self.iblock][1])
#         self.Base.cheml_type['regressor'].append(cheml_type)
#         dfx_train = self.type_check('dfx_train', cheml_type='df', req=True, py_type=pd.DataFrame).values
#         dfx_test = self.type_check('dfx_test', cheml_type='df', req=True, py_type=pd.DataFrame).values
#         dfy_train = self.type_check('dfy_train', cheml_type='df', req=True, py_type=pd.DataFrame)
#         dfy_header = dfy_train.columns
#         dfy_train = dfy_train.values
#         dfy_test = self.type_check('dfy_test', cheml_type='df', req=True, py_type=pd.DataFrame).values
#
#         try:
#             model = nn_psgd.train(dfx_train,dfx_test,dfy_train,dfy_test,**self.parameters)
#         except Exception as err:
#             msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
#                 err).__name__ + ': ' + str(err.message)
#             raise TypeError(msg)
#
#         dfy_pred = nn_psgd.output(dfx_train,model)
#         dfy_pred = pd.DataFrame(dfy_pred, columns=dfy_header)
#
#         order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
#         for token in set(order):
#             if token == 'model':
#                 self.Base.send[(self.iblock, token)] = [model, order.count(token),
#                                                         (self.iblock, token, self.Host, self.Function)]
#             elif token == 'dfy_train_pred':
#                 self.Base.send[(self.iblock, token)] = [dfy_pred, order.count(token),
#                                                         (self.iblock, token, self.Host, self.Function)]
#             else:
#                 msg = "@Task #%i(%s): non valid output token '%s'" % (self.iblock + 1, self.Task, token)
#                 raise NameError(msg)
#         del self.legal_inputs

##################################################################### 6 Mix

class kfold_pool(BASE):
    def legal_IO(self):
        self.legal_inputs = {'dfx': None, 'dfy': None, 'kfold':None, 'model':None, 'evaluator':None}
        self.legal_outputs = {'evaluation_results_': None, 'best_model_': None}
        requirements = ['scikit_learn']
        self.Base.requirements += [i for i in requirements if i not in self.Base.requirements]

    def fit(self):
        # step2: assign inputs to parameters if necessary (param = @token)
        self.paramFROMinput()

        # step4: import module and make APIs
        try:
            self._reg_evaluation_params()
        except Exception as err:
            msg = '@Task #%i(%s): '%(self.iblock+1, self.Task) + type(err).__name__ + ': '+ str(err.message)
            raise TypeError(msg)

        # step5: process
        # step6: send out
        order = [edge[1] for edge in self.Base.graph if edge[0]==self.iblock]
        for token in set(order):
            if token == 'evaluator':
                self.Base.send[(self.iblock, token)] = [self.evaluator, order.count(token),
                                                        (self.iblock, token, self.Host, self.Function)]
            elif token == 'evaluation_results_':
                # step1: check inputs
                dfy, dfy_info = self.input_check('dfy', req=True, py_type=pd.DataFrame)
                dfy_pred, dfy_pred_info = self.input_check('dfy_pred', req=True, py_type=pd.DataFrame)

                # step3: check the dimension of input data frame
                dfy, _ = self.data_check('dfy', dfy, ndim=2, n0=None, n1=None, format_out='df')
                dfy_pred, _ = self.data_check('dfy_pred', dfy_pred, ndim=2, n0=dfy.shape[0], n1=None, format_out='df')

                self._reg_evaluate(dfy, dfy_pred, self.evaluator)
                evaluation_results_ = self.results
                self.Base.send[(self.iblock, token)] = [pd.DataFrame(evaluation_results_), order.count(token),
                                                        (self.iblock,token,self.Host,self.Function)]
            else:
                msg = "@Task #%i(%s): non valid output token '%s'" % (self.iblock+1, self.Task, token)
                raise NameError(msg)

        #step7: delete all inputs from memory
        del self.legal_inputs

##################################################################### 7 Search
class GA_DEAP(BASE):
    def fit(self):
        # parameters
        self.paramFROMinput()
        # func_method default value
        if 'func_method' in self.parameters:
            method = self.parameters.pop('func_method')
        else:
            method = 'algorithm_1'
        if 'init_pop_frac' in self.parameters:
            init_pop_frac = self.parameters.pop('init_pop_frac')
        else:
            init_pop_frac = 0.35
        if 'crossover_pop_frac' in self.parameters:
            crossover_pop_frac = self.parameters.pop('crossover_pop_frac')
        else:
            crossover_pop_frac = 0.35

        # step4: import module and make APIs
        try:
            from cheml.search import GA_DEAP
            model = GA_DEAP(**self.parameters)
            model.fit()
            if method == 'algorithm_1':
                best_ind_df, best_individual = model.algorithm_1()
            elif method == 'algorithm_2':
                best_ind_df, best_individual = model.algorithm_2(init_pop_frac=init_pop_frac,
                                                                 crossover_pop_frac=crossover_pop_frac)
            elif method == 'algorithm_3':
                best_ind_df, best_individual = model.algorithm_3()
            elif method == 'algorithm_4':
                best_ind_df, best_individual = model.algorithm_4(crossover_pop_frac=crossover_pop_frac)
            else:
                msg = "@Task #%i(%s): the func_method is not valid. try one of these methods: ('algorithm_1', " \
                      "'algorithm_2', 'algorithm_3', 'algorithm_4')"% (self.iblock + 1, self.Task)
                raise Exception(msg)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(
                err).__name__ + ': ' + str(err.message)
            raise TypeError(msg)

        # step5: process
        # step6: send out
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'best_ind_df':
                self.set_value(token, best_ind_df)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]
            elif token == 'best_individual':
                removed_columns_ = pd.DataFrame(list(best_individual))
                self.set_value(token, removed_columns_)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        # step7: delete all inputs from memory
        del self.inputs



##################################################################### 8 Store

class SaveFile(BASE):
    def fit(self):
        # step1: check inputs
        self.required('df', req=True)
        df = self.inputs['df'].value

        # step2: assign inputs to parameters if necessary (param = @token)
        self.paramFROMinput()

        # step3: check the dimension of input data frame
        # step4: import module and make APIs
        try:
            from cheml.initialization import SaveFile
            model = SaveFile(**self.parameters)
            model.fit(df, self.Base.output_directory)
        except Exception as err:
            msg = '@Task #%i(%s): '%(self.iblock+1, self.Task) + type(err).__name__ + ': '+ str(err.message)
            raise TypeError(msg)

        # step5: process
        # step6: send out
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token not in self.outputs:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)
            elif token == 'filepath':
                self.set_value(token, model.file_path)
                self.outputs[token].count = order.count(token)
                self.Base.send[(self.iblock, token)] = self.outputs[token]

        # step7: delete all inputs from memory
        del self.inputs





import datetime
import numpy as np
import sys

def list_del_indices(mylist,indices):
    for index in sorted(indices, reverse=True):
        del mylist[index]
    return mylist

def std_datetime_str(mode='datetime'):
    """(std_time_str):
        This function gives out the formatted time as a standard string, i.e., YYYY-MM-DD hh:mm:ss.
    """
    if mode == 'datetime':
        return str(datetime.datetime.now())[:19]
    elif mode == 'date':
        return str(datetime.datetime.now())[:10]
    elif mode == 'time':
        return str(datetime.datetime.now())[11:19]
    elif mode == 'datetime_ms':
        return str(datetime.datetime.now())
    elif mode == 'time_ms':
        return str(datetime.datetime.now())[11:]
    else:
        sys.exit("Invalid mode!")

def slurm_script_exclusive(pyscript_file,nnodes=1,input_slurm_script=None,output_slurm_script='script.slurm'):
    """(slurmjob)
    make the slurm script based on exclusive selection of cores per nodes.

    Parameters
    ----------
    pyscript_file: string
        This is the python script that includes nn_dsgd functions and you are 
        going to run on the cluster. If you are using the cheml python script
        maker this parameter is going to be the name of the final output file.
         
    nnodes: int, optional(default = 1) 
        number of available empty nodes in the cluster.
    
    input_slurm_script: string, optional(default = None)
        The file path to the prepared slurm script. We also locate place of
        --nodes and -np in the script and make sure that provided numbers are 
        equal to number of nodes(nnodes). Also, the exclusive option must be 
        included in the script to have access to an entire node.
    
    output_slurm_script: string, optional(default = 'script.slurm')
        The path and name of the slurm script file that will be saved after 
        changes by this function.
        
    Returns
    -------
    The function will write a slurm script file with the filename passed by 
    output_slurm_script.
    """
    if not input_slurm_script:
        file = ['#!/bin/sh\n', '#SBATCH --time=99:00:00\n', '#SBATCH --job-name="nn"\n', '#SBATCH --output=nn.out\n', '#SBATCH --clusters=chemistry\n', '#SBATCH --partition=beta\n', '#SBATCH --account=pi-hachmann\n', '#SBATCH --exclusive\n', '#SBATCH --nodes=1\n', '\n', '# ====================================================\n', '# For 16-core nodes\n', '# ====================================================\n', '#SBATCH --constraint=CPU-E5-2630v3\n', '#SBATCH --tasks-per-node=1\n', '#SBATCH --mem=64000\n', '\n', '\n', 'echo "SLURM job ID         = "$SLURM_JOB_ID\n', 'echo "Working Dir          = "$SLURM_SUBMIT_DIR\n', 'echo "Temporary scratch    = "$SLURMTMPDIR\n', 'echo "Compute Nodes        = "$SLURM_NODELIST\n', 'echo "Number of Processors = "$SLURM_NPROCS\n', 'echo "Number of Nodes      = "$SLURM_NNODES\n', 'echo "Tasks per Node       = "$TPN\n', 'echo "Memory per Node      = "$SLURM_MEM_PER_NODE\n', '\n', 'ulimit -s unlimited\n', 'module load intel-mpi\n', 'module load python\n', 'module list\n', 'export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:/projects/hachmann/packages/Anaconda:/projects/hachmann/packages/rdkit-Release_2015_03_1:/user/m27/pkg/openbabel/2.3.2/lib\n', 'date\n', '\n', '\n', 'echo "Launch job"\n', 'export I_MPI_PMI_LIBRARY=/usr/lib64/libpmi.so\n', 'export I_MPI_FABRICS=shm:tcp\n', '\n', 'mpirun -np 2 python test.py\n']
        file[8] = '#SBATCH --nodes=%i\n'%nnodes
        file[-1] = 'mpirun -np %i python %s\n' %(nnodes,pyscript_file)
    else:
        file = open(input_slurm_script,'r')
        file = file.readlines()
        exclusive_flag = False
        nodes_flag = False
        np_flag = False
        for i,line in enumerate(file):
            if '--exclusive' in line:
                exclusive_flag = True
            elif '--nodes' in line:
                nodes_flag = True
                ind = line.index('--nodes')
                file[i] = line[:ind]+'--nodes=%i\n'%nnodes
            elif '-np' in line:
                np_flag = True
                ind = line.index('--nodes')
                file[i] = line[:ind]+'--nodes=%i\n'%nnodes                     
        if not exclusive_flag:
            file = file[0] + ['#SBATCH --exclusive\n'] + file[1:]
            msg = "The --exclusive option is not available in the slurm script. We added '#SBATCH --exclusive' to the first of file."
            warnings.warn(msg,UserWarning)
        if not nodes_flag:
            file = file[0] + ['#SBATCH --nodes=%i\n'%nnodes] + file[1:]
            msg = "The --nodes option is not available in the slurm script. We added '#SBATCH --nodes=%i' to the first of file."%nnodes
            warnings.warn(msg,UserWarning)
        if not np_flag:
            file.append('mpirun -np %i python %s\n'%(nnodes,pyscript_file))
            msg = "The -np option is not available in the slurm script. We added 'mpirun -np %i python %s'to the end of file."%(nnodes,pyscript_file) 
            warnings.warn(msg,UserWarning)
            
    script = open(output_slurm_script,'w')
    for line in file:    
        script.write(line)
    script.close()

def chunk(xs, n, X=None, Y=None):
    """
    X and Y must be np array
    the term of use:
                it = chunk ( range(len(X), n, X, Y)
                X_chunk, y_chunk = it.next()
    n is the number of chunks (#total_batch).
    """
    ys = list(xs)
    # random.shuffle(ys)
    size = len(ys) // n
    leftovers= ys[size*n:]
    for c in xrange(n):
        if leftovers:
           extra= [ leftovers.pop() ] 
        else:
           extra= []
        if isinstance(X,np.ndarray):
            if isinstance(Y, np.ndarray):
                yield X[ys[c*size:(c+1)*size] + extra], Y[ys[c*size:(c+1)*size] + extra]
            else:
                yield X[ys[c * size:(c + 1) * size] + extra]
        else:
            yield ys[c*size:(c+1)*size] + extra

def choice(X, Y=None, n=0.1, replace=False):
    """
    Generates a random sample from a given 1-D array. Bassicaly same as np.random.choice with pre- and post-processing
     steps. Sampling without replacement.
    :param X: 1-D array-like
        A random sample will be generated from its elements.
    :param Y: 1-D array-like, optional (default = None)
        A random sample will be generated from its elements.
    :param n: int or float between zero and one, optional (default = 0.1)
        size of sample
    :param replace: boolean, default=False
        whether the sample is with or without replacement
    :return: a_out: 1-D array-like
                the array of out of sample elements
    :return: a_sample: 1-D array-like, shape (size,)
                the sample array
    """
    if not isinstance(n,int):
            n = int(n*len(X))
    ind_sample = np.random.choice(len(X),n,replace=replace)
    ind_out = np.array([i for i in xrange(len(X)) if i not in ind_sample])
    X_sample = X[ind_sample]
    X_out = X[ind_out]
    if isinstance(Y,np.ndarray):
        if len(Y) != len(X):
            raise Exception('X and Y must be same size')
        Y_sample = Y[ind_sample]
        Y_out = Y[ind_out]
    else:
        Y_sample = None
        Y_out = None
    return X_out, X_sample, Y_out, Y_sample

def return2Dshape(shape):
    if len(shape) == 2:
        return shape
    elif len(shape) == 1:
        return (shape[0],None)
    else:
        raise Exception('input dimension is greater than 2')

def bool_formatter(bool):
    if bool:
        return("true")
    else:
        return("false")



import unittest
import os
import pkg_resources
import pandas as pd
import numpy as np



DATA_PATH = pkg_resources.resource_filename('cheml', os.path.join('tests', 'data'))

from cheml.preprocessing import MissingValues


# dummy data
df = pd.read_csv(os.path.join(DATA_PATH,'test_missing_values.csv'), header=None)
target=pd.DataFrame([1,2,3,np.nan,4])


class TestConstantColumns(unittest.TestCase):
    def test_zero(self):
        mv = MissingValues(strategy = 'zero',
                           string_as_null = True,
                           inf_as_null = True,
                           missing_values = None)
        f = mv.fit_transform(df)
        t = mv.fit_transform(target)
        self.assertEqual((5,9), f.shape)
        self.assertEqual(0.0, f[10][0])
        self.assertEqual(0.0, f[10][2])
        self.assertEqual(0.0, f[2][1])
        self.assertEqual(0.0, f[1][0])

if __name__== '__main__':
    unittest.main()




import unittest
import pandas as pd

from cheml.preprocessing import Outliers


# dummy data
df = pd.DataFrame([1,2,3])
df[1] = [1,1.1,9]
df[2] = [7]*3
df[3] = [5.432423]*3


class TestOutliers(unittest.TestCase):
    def test_mean(self):
        ro = Outliers(m=1.3,strategy='mean')
        f = ro.fit_transform(df)
        self.assertEqual(f.shape, (2, 4))
        self.assertEqual(1, f.index[1])

    def test_median(self):
        ro = Outliers(m=1.,strategy='median')
        f = ro.fit_transform(df)
        ff = ro.transform(df)
        self.assertEqual(f.shape, (1, 4))
        self.assertEqual(1, f.index[0])
        self.assertEqual(1, (ff==f).sum()[0])

if __name__== '__main__':
    unittest.main()




from deap import base, creator, tools
import random
import pandas as pd
import time
import math

class GA_DEAP(object):
    """
    A genetic algorithm class for search or optimization problems, built on top of the
    Distributed Evolutionary Algorithms in Python (DEAP) library. There are four algorithms with different genetic
    algorithm selection methods. The methods are described individually in each algorithm's documentation.

    Parameters
    ----------
    Evaluate: function
        The objective function that has to be optimized. The first parameter of the objective function should be
        an object of type <chromosome_type>. Objective function should return a tuple

    Weights: tuple of integer(s), optional (default = (-1.0, ) )
        A tuple containing fitness objective(s) for objective function(s). Ex: (1.0,) for maximizing and (-1.0,)
        for minimizing a single objective function

    chromosome_length: integer, optional (default = 1)
        An integer which specifies the length of chromosome/individual

    chromosome_type: tuple of <chromosome_length> integers, optional (default = (1,) )
        A tuple of integers of length <chromosome_length> describing the type of each bit of the chromosome.
        0 for floating type and 1 for integer type. All integer types first followed by the floating types

    bit_limits: tuple of <chromosome_length> tuples, optional (default = (0,10) )
        A tuple of <chromosome_length> tuples containing the lower and upper bounds for each bit of individual

    pop_size: integer, optional (default = 50)
        An integer which denotes the size of the population

    crossover_prob: real number, optional (default = 0.4)
        A number that denotes the probability of crossover

    crossover_type: string, optional (default = "Blend")
        A string denoting the type of crossover: OnePoint, TwoPoint, Blend or Uniform

    mutation_prob: real number, optional (default = 0.4)
        A number that denotes the probability of mutation.

    mut_float_mean: real number, optional (default = 0)
        Value of the mean of the Gaussian distribution for Gaussian type mutation

    mut_float_dev: real number, optional (default = 1)
        Value of the standard deviation of the Gaussian distribution for Gaussian type mutation

    mut_int_lower: tuple of integers, optional (default = 1)
        A tuple of integers of length (total number of integers in the chromosome) containing lower limit(s)
        (inclusive) for integer type mutation

    mut_int_upper: tuple of integers, optional (default = 10)
        A tuple of integers of length (total number of integers in the chromosome) containing upper limit(s)
        (inclusive) for integer type mutation

    n_generations: integer, optional (default = 20)
        An integer for the number of generations for evolving the population

    Examples
    --------
    >>> from cheml.search import GA_DEAP
    >>> def sum_func(individual): return (sum(individual),)
    >>> ga_search = GA_DEAP(Evaluate = sum_func, Weights = (1,), chromosome_length = 2, chromosome_type = (1,1),
    >>>       bit_limits = ((0,10), (0,5)), mut_int_lower = (0,0), mut_int_upper = (10,5))
    >>> ga_search.fit()
    >>> best_ind_df, best_individual = ga_search.algorithm_1()
    Best Individuals of each generation are:
       Best_individual_per_gen  Fitness_values          Time
    0                   [8, 5]            13.0  4.444453e-06
    1                  [10, 4]            14.0  5.000035e-06
    2                  [10, 4]            14.0  0.000000e+00
    3                  [10, 4]            14.0  4.166696e-06
    4                  [10, 4]            14.0  4.444387e-06
    5                  [10, 5]            15.0  4.444453e-06
    6                  [10, 5]            15.0  4.166696e-06
    7                  [10, 5]            15.0  4.444453e-06
    8                  [10, 5]            15.0  0.000000e+00
    9                  [10, 5]            15.0  8.333392e-07
    10                 [10, 5]            15.0  4.444453e-06
    11                 [10, 5]            15.0  4.166696e-06
    12                 [10, 5]            15.0  4.444453e-06
    13                 [10, 5]            15.0  4.444453e-06
    14                 [10, 5]            15.0  0.000000e+00
    15                 [10, 5]            15.0  0.000000e+00
    16                 [10, 5]            15.0  8.333392e-07
    17                 [10, 5]            15.0  4.444453e-06
    18                 [10, 5]            15.0  4.444453e-06
    19                 [10, 5]            15.0  4.166630e-06
    ((0, 10), (0, 5))
     Best individual after 20 evolutions is [10, 5]
    """
    def __init__(self, Evaluate, Weights = (-1.0,), chromosome_length = 1, chromosome_type = (1,), bit_limits = ((0,10),), \
                 pop_size = 50, crossover_prob = 0.4, crossover_type = "Blend", mutation_prob = 0.4, \
                 mut_float_mean = 0, mut_float_dev = 1, mut_int_lower = (1,), mut_int_upper = (10,), n_generations = 20):

        self.Weights = Weights
        self.chromosome_length = chromosome_length
        self.chromosome_type = chromosome_type
        self.bit_limits = bit_limits
        self.evaluate = Evaluate
        self.pop_size = pop_size
        self.crossover_prob = crossover_prob
        self.mutation_prob = mutation_prob
        self.crossover_type = crossover_type
        self.mut_float_param_1 = mut_float_mean
        self.mut_float_param_2 = mut_float_dev
        self.mut_int_param_1 = mut_int_lower
        self.mut_int_param_2 = mut_int_upper
        self.n_generations = n_generations
        self.n_integers = 0
        for i in chromosome_type:
            if i == 1:
                self.n_integers += 1


    def chromosome_generator(self):
        A = []
        for i in range(self.chromosome_length):
            if self.chromosome_length == 1:
                if self.chromosome_type[i] == 0:
                    A.append(random.uniform(self.bit_limits[0], self.bit_limits[1]))
                else:
                    A.append(random.randint(self.bit_limits[0], self.bit_limits[1]))
            else:
                if self.chromosome_type[i] == 0:
                    A.append(random.uniform(self.bit_limits[i][0], self.bit_limits[i][1]))
                else:
                    A.append(random.randint(self.bit_limits[i][0], self.bit_limits[i][1]))
        return A


    def fit(self):
        """
        Setting up the DEAP - genetic algorithm parameters and functions.

        Notes
        -----
        Must call this function before calling any algorithm.
        """
        creator.create("FitnessMin", base.Fitness, weights=self.Weights)
        creator.create("Individual", list, fitness=creator.FitnessMin)
        self.toolbox = base.Toolbox()
        self.toolbox.register("individual", tools.initIterate, creator.Individual, self.chromosome_generator)
        self.toolbox.register("population", tools.initRepeat, list, self.toolbox.individual)

        if self.crossover_type == "OnePoint":
            self.toolbox.register("mate", tools.cxOnePoint)
        elif self.crossover_type == "TwoPoint":
            self.toolbox.register("mate", tools.cxTwoPoint)
        elif self.crossover_type == "Uniform":
            self.toolbox.register("mate", tools.cxUniform, indpb=self.crossover_prob)
        elif self.crossover_type == "Blend":
            self.toolbox.register("mate", tools.cxBlend, alpha = 0.5)

        self.toolbox.register("selectTournament", tools.selTournament, tournsize=30)
        self.toolbox.register("selectRoulette", tools.selRoulette)

        def feasibility(indi):
            for x,i in zip(indi,range(self.chromosome_length)):
                if self.chromosome_length == 1:
                    if self.bit_limits[0] <= x <= self.bit_limits[1]:
                        continue
                    else:
                        return False
                else:
                    if self.bit_limits[i][0] <= x <= self.bit_limits[i][1]:
                        continue
                    else:
                        return False
            return True

        def distance(indi):
            s = 0
            for x,i in zip(indi, range(self.chromosome_length)):
                if self.chromosome_length == 1:
                    low = self.bit_limits[0]
                    up = self.bit_limits[1]
                else:
                    low = self.bit_limits[i][0]
                    up = self.bit_limits[i][1]
                if x < low:
                    s += ((x-low)*100)**2
                elif x > up:
                    s += ((x-up)*100)**2
            return s

        self.toolbox.register("evaluate", self.evaluate)
        self.toolbox.decorate("evaluate", tools.DeltaPenalty(feasibility, -100.0*self.Weights[0], distance))


    def custom_mutate(self, indi):
        for i in range(self.chromosome_length):
            if self.chromosome_type[i] == 0:
                if random.random() < self.mutation_prob:
                    indi[i] += random.gauss(self.mut_float_param_1, self.mut_float_param_2)
            elif self.chromosome_type[i] == 1:
                if random.random() < self.mutation_prob:
                    indi[i] = random.randint(self.mut_int_param_1[i], self.mut_int_param_2[i])


    def algorithm_1(self):
        """
        Initial population is instantiated.
        Roulette wheel selection is used for selecting individuals for crossover and mutation.
        The initial population, crossovered and mutated individuals form the pool of individuals from which the best -
        n members are selected as the initial population for the next generation, where n is the size of population.

        Returns
        -------
        best_ind_df:  pandas dataframe
            A pandas dataframe of best individuals of each generation

        best_ind:  list of <chromosome_length> numbers
            The best individual after the last generation.
        Notes
        -----
        """

        pop = self.toolbox.population(n=self.pop_size)
        # Evaluate the entire population
        fitnesses = map(self.toolbox.evaluate, pop)
        for ind, fit in zip(pop, fitnesses):
            ind.fitness.values = fit
        co_pop = self.toolbox.selectRoulette(pop, int(math.ceil(0.8*len(pop))))
        co_pop = list(map(self.toolbox.clone, co_pop))
        mu_pop = self.toolbox.selectRoulette(pop, int(math.ceil(0.3*len(pop))))
        mu_pop = list(map(self.toolbox.clone, mu_pop))

        for child1, child2 in zip(co_pop[::2], co_pop[1::2]):
            self.toolbox.mate(child1, child2)
            for i in range(self.chromosome_length):
                if self.chromosome_type[i] == 1:
                    child1[i] = int(child1[i])
                    child2[i] = int(child2[i])
            del child1.fitness.values
            del child2.fitness.values

        for mutant in mu_pop:
            self.custom_mutate(mutant)
            del mutant.fitness.values

        # Evaluate the crossover and mutated population
        total_pop = pop + co_pop + mu_pop
        invalid_ind = [ind for ind in total_pop if not ind.fitness.valid]
        fitnesses = list(map(self.toolbox.evaluate, invalid_ind))
        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit

#        fits = [indi.fitness.values[0] for indi in pop]

        best_indi_per_gen = []
        best_indi_fitness_values = []
        timer = []
        for g in range(self.n_generations):
            st_time = time.time()
            # Select the next generation individuals
            offspring = tools.selBest(total_pop, self.pop_size)
            # Clone the selected individuals
            offspring = list(map(self.toolbox.clone, offspring))
            co_pop = self.toolbox.selectRoulette(offspring, int(math.ceil(0.8*len(pop))))
            co_pop = list(map(self.toolbox.clone, co_pop))
            mu_pop = self.toolbox.selectRoulette(offspring, int(math.ceil(0.3*len(pop))))
            mu_pop = list(map(self.toolbox.clone, mu_pop))

            for child1, child2 in zip(co_pop[::2], co_pop[1::2]):
                self.toolbox.mate(child1, child2)
                for i in range(self.chromosome_length):
                    if self.chromosome_type[i] == 1:
                        child1[i] = int(child1[i])
                        child2[i] = int(child2[i])
                del child1.fitness.values
                del child2.fitness.values

            for mutant in mu_pop:
                self.custom_mutate(mutant)
                del mutant.fitness.values

            # Evaluate the crossover and mutated population
            total_pop = offspring + co_pop + mu_pop
            invalid_ind = [ind for ind in total_pop if not ind.fitness.valid]
            fitnesses = list(map(self.toolbox.evaluate, invalid_ind))
            for ind, fit in zip(invalid_ind, fitnesses):
                ind.fitness.values = fit


            # Storing the best individuals after each generation
            best_individual = tools.selBest(total_pop, 1)[0]
            best_indi_per_gen.append(list(best_individual))
            best_indi_fitness_values.append(best_individual.fitness.values[0])


            tot_time = (time.time() - st_time)/(60*60)
            timer.append(tot_time)

            b1 = pd.Series(best_indi_per_gen, name = 'Best_individual_per_gen')
            b2 = pd.Series(best_indi_fitness_values, name = 'Fitness_values')
            b3 = pd.Series(timer, name = 'Time')
            best_ind_df = pd.concat([b1,b2,b3], axis=1)
            # best_ind_df.to_csv('best_ind.csv',index=False)

        # best_ind_df = pd.DataFrame(best_indi_per_gen)
        print "\n \n Best Individuals of each generation are:  \n \n" , best_ind_df
        print "\n \n" , self.bit_limits, " \n \n Best individual after %s evolutions is %s " % (self.n_generations, best_individual)
        return best_ind_df, best_individual


    def algorithm_2(self, init_pop_frac = 0.35, crossover_pop_frac = 0.35):
        """
        Initial population is instantiated.
        Roulette wheel selection is used for selecting individuals for crossover and mutation.
        The initial population, crossovered and mutated individuals form 3 different pools of individuals. Based on
        input parameters 1 and 2, members are selected from each of these pools to form the initial population for the
        next generation. Fraction of mutated members to select for next generation is decided based on the 2 input
        parameters and the size of initial population.

        Parameters
        ----------
        init_pop_frac: float, optional (default = 0.4)
            Fraction of initial population to select for next generation

        crossover_pop_frac: float, optional (default = 0.3)
            Fraction of crossover population to select for next generation

        Returns
        -------
        best_ind_df:  pandas dataframe
            A pandas dataframe of best individuals of each generation

        best_ind:  list of <chromosome_length> numbers
            The best individual after the last generation.
        """
        pop = self.toolbox.population(n=self.pop_size)
        # Evaluate the entire population
        fitnesses = map(self.toolbox.evaluate, pop)
        for ind, fit in zip(pop, fitnesses):
            ind.fitness.values = fit
        co_pop = self.toolbox.selectRoulette(pop, int(math.ceil(0.8*len(pop))))
        co_pop = list(map(self.toolbox.clone, co_pop))
        mu_pop = self.toolbox.selectRoulette(pop, int(math.ceil(0.3*len(pop))))
        mu_pop = list(map(self.toolbox.clone, mu_pop))

        for child1, child2 in zip(co_pop[::2], co_pop[1::2]):
            self.toolbox.mate(child1, child2)
            for i in range(self.chromosome_length):
                if self.chromosome_type[i] == 1:
                    child1[i] = int(child1[i])
                    child2[i] = int(child2[i])
            del child1.fitness.values
            del child2.fitness.values

        for mutant in mu_pop:
            self.custom_mutate(mutant)
            del mutant.fitness.values

        # Evaluate the crossover and mutated population
        total_pop = pop + co_pop + mu_pop
        invalid_ind = [ind for ind in total_pop if not ind.fitness.valid]
        fitnesses = list(map(self.toolbox.evaluate, invalid_ind))
        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit

#        fits = [indi.fitness.values[0] for indi in pop]

        best_indi_per_gen = []
        best_indi_fitness_values = []
        timer = []
        for g in range(self.n_generations):
            st_time = time.time()
            # Select the next generation individuals
            off_pop = tools.selBest(pop, int(math.ceil(init_pop_frac*len(pop))))
            off_co = tools.selBest(co_pop, int(math.ceil(crossover_pop_frac*len(pop))))
            off_mu = tools.selBest(mu_pop, len(pop)-len(off_pop)-len(off_co) )
            offspring = off_pop + off_co + off_mu
            # Clone the selected individuals
            offspring = list(map(self.toolbox.clone, offspring))
            co_pop = self.toolbox.selectRoulette(offspring, int(math.ceil(0.8*len(pop))))
            co_pop = list(map(self.toolbox.clone, co_pop))
            mu_pop = self.toolbox.selectRoulette(offspring, int(math.ceil(0.3*len(pop))))
            mu_pop = list(map(self.toolbox.clone, mu_pop))

            for child1, child2 in zip(co_pop[::2], co_pop[1::2]):
                self.toolbox.mate(child1, child2)
                for i in range(self.chromosome_length):
                    if self.chromosome_type[i] == 1:
                        child1[i] = int(child1[i])
                        child2[i] = int(child2[i])
                del child1.fitness.values
                del child2.fitness.values

            for mutant in mu_pop:
                self.custom_mutate(mutant)
                del mutant.fitness.values

            # Evaluate the crossover and mutated population
            total_pop = offspring + co_pop + mu_pop
            invalid_ind = [ind for ind in total_pop if not ind.fitness.valid]
            fitnesses = list(map(self.toolbox.evaluate, invalid_ind))
            for ind, fit in zip(invalid_ind, fitnesses):
                ind.fitness.values = fit
            pop = offspring

            # Storing the best individuals after each generation
            best_individual = tools.selBest(total_pop, 1)[0]
            best_indi_per_gen.append(list(best_individual))
            best_indi_fitness_values.append(best_individual.fitness.values[0])


            tot_time = (time.time() - st_time)/(60*60)
            timer.append(tot_time)

            b1 = pd.Series(best_indi_per_gen, name = 'Best_individual_per_gen')
            b2 = pd.Series(best_indi_fitness_values, name = 'Fitness_values')
            b3 = pd.Series(timer, name = 'Time')
            best_ind_df = pd.concat([b1,b2,b3], axis=1)
            # best_ind_df.to_csv('best_ind.csv',index=False)


    #	best_ind_df = pd.DataFrame(best_indi_per_gen)
        print "\n \n Best Individuals of each generation are:  \n \n" , best_ind_df
        print "\n \n" , self.bit_limits, " \n \n Best individual after %s evolutions is %s " % (self.n_generations, best_individual)
        return best_ind_df, best_individual


    def algorithm_3(self):
        """
        Initial population is instantiated.
        Roulette wheel selection is used for selecting individuals for crossover and mutation.
        The initial population, crossovered and mutated individuals form the pool of individuals from which n members
        are selected using Roulette wheel selection, but without replacement to ensure uniqueness of members in the next
        generation, as the initial population for the next generation, where n is the size of population.

        Returns
        -------
        best_ind_df:  pandas dataframe
            A pandas dataframe of best individuals of each generation

        best_ind:  list of <chromosome_length> numbers
            The best individual after the last generation.
        """
        pop = self.toolbox.population(n=self.pop_size)
        # Evaluate the entire population
        fitnesses = map(self.toolbox.evaluate, pop)
        for ind, fit in zip(pop, fitnesses):
            ind.fitness.values = fit
        co_pop = self.toolbox.selectRoulette(pop, int(math.ceil(0.8*len(pop))))
        co_pop = list(map(self.toolbox.clone, co_pop))
        mu_pop = self.toolbox.selectRoulette(pop, int(math.ceil(0.3*len(pop))))
        mu_pop = list(map(self.toolbox.clone, mu_pop))

        for child1, child2 in zip(co_pop[::2], co_pop[1::2]):
            self.toolbox.mate(child1, child2)
            for i in range(self.chromosome_length):
                if self.chromosome_type[i] == 1:
                    child1[i] = int(child1[i])
                    child2[i] = int(child2[i])
            del child1.fitness.values
            del child2.fitness.values

        for mutant in mu_pop:
            self.custom_mutate(mutant)
            del mutant.fitness.values

            # Evaluate the crossover and mutated population
        total_pop = pop + co_pop + mu_pop
        invalid_ind = [ind for ind in total_pop if not ind.fitness.valid]
        fitnesses = list(map(self.toolbox.evaluate, invalid_ind))
        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit

#        fits = [indi.fitness.values[0] for indi in pop]

        best_indi_per_gen = []
        best_indi_fitness_values = []
        timer = []
        for g in range(self.n_generations):
            st_time = time.time()
            # Select the next generation individuals
            #offspring = tools.selBest(total_pop, self.pop_size)
            #temp_pop = total_pop
            offspring = []
            for i in range(len(pop)):
                #off = self.toolbox.selectRoulette(total_pop, 1)
                off = tools.selBest(total_pop, 1)
                offspring.append(off[0])
                total_pop.remove(off[0])
            # Clone the selected individuals
            # offspring = list(map(self.toolbox.clone, offspring))

            co_pop = self.toolbox.selectRoulette(offspring, int(math.ceil(0.8*len(pop))))
            co_pop = list(map(self.toolbox.clone, co_pop))
            mu_pop = self.toolbox.selectRoulette(offspring, int(math.ceil(0.3*len(pop))))
            mu_pop = list(map(self.toolbox.clone, mu_pop))

            for child1, child2 in zip(co_pop[::2], co_pop[1::2]):
                self.toolbox.mate(child1, child2)
                for i in range(self.chromosome_length):
                    if self.chromosome_type[i] == 1:
                        child1[i] = int(child1[i])
                        child2[i] = int(child2[i])
                del child1.fitness.values
                del child2.fitness.values

            for mutant in mu_pop:
                self.custom_mutate(mutant)
                del mutant.fitness.values

            # Evaluate the crossover and mutated population
            total_pop = offspring + co_pop + mu_pop
            invalid_ind = [ind for ind in total_pop if not ind.fitness.valid]
            fitnesses = list(map(self.toolbox.evaluate, invalid_ind))
            for ind, fit in zip(invalid_ind, fitnesses):
                ind.fitness.values = fit


            # Storing the best individuals after each generation
            best_individual = tools.selBest(total_pop, 1)[0]
            best_indi_per_gen.append(list(best_individual))
            best_indi_fitness_values.append(best_individual.fitness.values[0])

            tot_time = (time.time() - st_time)/(60*60)
            timer.append(tot_time)
            b1 = pd.Series(best_indi_per_gen, name = 'Best_individual_per_gen')
            b2 = pd.Series(best_indi_fitness_values, name = 'Fitness_values')
            b3 = pd.Series(timer, name = 'Time')
            best_ind_df = pd.concat([b1,b2,b3], axis=1)
            # best_ind_df.to_csv('best_ind.csv',index=False)

    #	best_ind_df = pd.DataFrame(best_indi_per_gen)
        print "\n \n Best Individuals of each generation are:  \n \n" , best_ind_df
        print "\n \n" , self.bit_limits, " \n \n Best individual after %s evolutions is %s " % (self.n_generations, best_individual)
        return best_ind_df, best_individual


    def algorithm_4(self, crossover_pop_frac = 0.4):
        """
        Initial population is instantiated.
        Roulette wheel selection is used for selecting individuals for crossover. For mutation, fraction of individuals
        are selected from the crossover population based on input parameter.
        The initial population, crossovered and mutated individuals form the pool of individuals from which the best -
        n members are selected as the initial population for the next generation, where n is the size of population.

        Parameters
        ----------
        crossover_pop_frac: float, optional (default = 0.4)
            Fraction of crossover population to select for mutation

        Returns
        -------
        best_ind_df:  pandas dataframe
            A pandas dataframe of best individuals of each generation

        best_ind:  list of <chromosome_length> numbers
            The best individual after the last generation.
        """
        pop = self.toolbox.population(n=self.pop_size)
        # Evaluate the entire population
        fitnesses = map(self.toolbox.evaluate, pop)
        for ind, fit in zip(pop, fitnesses):
            ind.fitness.values = fit
        co_pop = self.toolbox.selectRoulette(pop, int(math.ceil(0.8*len(pop))))
        co_pop = list(map(self.toolbox.clone, co_pop))
        mu_pop = self.toolbox.selectRoulette(co_pop, int(math.ceil(crossover_pop_frac*len(co_pop))))
        mu_pop = list(map(self.toolbox.clone, mu_pop))
        for i in mu_pop:
            if i in co_pop:
                co_pop.remove(i)

        for child1, child2 in zip(co_pop[::2], co_pop[1::2]):
            self.toolbox.mate(child1, child2)
            for i in range(self.chromosome_length):
                if self.chromosome_type[i] == 1:
                    child1[i] = int(child1[i])
                    child2[i] = int(child2[i])
            del child1.fitness.values
            del child2.fitness.values

        for mutant in mu_pop:
            self.custom_mutate(mutant)
            del mutant.fitness.values

        # Evaluate the crossover and mutated population
        total_pop = pop + co_pop + mu_pop
        invalid_ind = [ind for ind in total_pop if not ind.fitness.valid]
        fitnesses = list(map(self.toolbox.evaluate, invalid_ind))
        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit

#        fits = [indi.fitness.values[0] for indi in pop]

        best_indi_per_gen = []
        best_indi_fitness_values = []
        timer = []
        for g in range(self.n_generations):
            st_time = time.time()
                # Select the next generation individuals
            offspring = tools.selBest(total_pop, self.pop_size)
            # Clone the selected individuals
            offspring = list(map(self.toolbox.clone, offspring))
            co_pop = self.toolbox.selectRoulette(offspring, int(math.ceil(0.8*len(pop))))
            co_pop = list(map(self.toolbox.clone, co_pop))
            mu_pop = self.toolbox.selectRoulette(co_pop, int(math.ceil(crossover_pop_frac*len(co_pop))))
            mu_pop = list(map(self.toolbox.clone, mu_pop))
            for i in mu_pop:
                if i in co_pop:
                    co_pop.remove(i)


            for child1, child2 in zip(co_pop[::2], co_pop[1::2]):
                self.toolbox.mate(child1, child2)
                for i in range(self.chromosome_length):
                    if self.chromosome_type[i] == 1:
                        child1[i] = int(child1[i])
                        child2[i] = int(child2[i])
                del child1.fitness.values
                del child2.fitness.values

            for mutant in mu_pop:
                self.custom_mutate(mutant)
                del mutant.fitness.values

                # Evaluate the crossover and mutated population
            total_pop = offspring + co_pop + mu_pop
            invalid_ind = [ind for ind in total_pop if not ind.fitness.valid]
            fitnesses = list(map(self.toolbox.evaluate, invalid_ind))
            for ind, fit in zip(invalid_ind, fitnesses):
                ind.fitness.values = fit


            # Storing the best individuals after each generation
            best_individual = tools.selBest(total_pop, 1)[0]
            best_indi_per_gen.append(list(best_individual))
            best_indi_fitness_values.append(best_individual.fitness.values[0])


            tot_time = (time.time() - st_time)/(60*60)
            timer.append(tot_time)

            b1 = pd.Series(best_indi_per_gen, name = 'Best_individual_per_gen')
            b2 = pd.Series(best_indi_fitness_values, name = 'Fitness_values')
            b3 = pd.Series(timer, name = 'Time')
            best_ind_df = pd.concat([b1,b2,b3], axis=1)
            # best_ind_df.to_csv('best_ind.csv',index=False)

    #	best_ind_df = pd.DataFrame(best_indi_per_gen)
        print "\n \n Best Individuals of each generation are:  \n \n" , best_ind_df
        print "\n \n" , self.bit_limits, " \n \n Best individual after %s evolutions is %s " % (self.n_generations, best_individual)
        return best_ind_df, best_individual




import unittest
from .....vassal.analysis.voronoi.VoronoiCell import VoronoiCell
from .....vassal.analysis.voronoi.VoronoiFace import VoronoiFace
from .....vassal.data.Atom import Atom
from .....vassal.data.AtomImage import AtomImage
from .....vassal.data.Cell import Cell
from .....vassal.geometry.Plane import Plane
from .....vassal.util.VectorCombinationComputer import \
    VectorCombinationComputer

class testVoronoiCell(unittest.TestCase):
    def test_supercell(self):
        # Create cell.
        structure = Cell()
        structure.add_atom(Atom([0, 0, 0], 0))

        # Create cell for atom1.
        cell = VoronoiCell(structure.get_atom(0), radical=True)

        # Compute faces.
        images = [AtomImage(structure.get_atom(0), sc) for sc in
                  VectorCombinationComputer(
                    structure.get_lattice_vectors(),
                      1.1).get_supercell_coordinates()]
        faces = cell.compute_faces(images)

        # Get direct neighbors.
        direct_faces = cell.compute_direct_neighbors(faces)

        # Simple tests.
        self.assertEquals(6, len(direct_faces))
        self.assertEquals(len(images) - 6 - 1, len(faces))

        # Make sure direct faces match up with expectations.
        neighboring_faces = []
        neighboring_faces.append(VoronoiFace(structure.get_atom(0),
                                             AtomImage(structure.get_atom(0),
                                                       [1, 0, 0]),
                                             radical=True))
        neighboring_faces.append(VoronoiFace(structure.get_atom(0),
                                             AtomImage(structure.get_atom(0),
                                                       [-1, 0, 0]),
                                             radical=True))
        neighboring_faces.append(VoronoiFace(structure.get_atom(0),
                                             AtomImage(structure.get_atom(0),
                                                       [0, 1, 0]),
                                             radical=True))
        neighboring_faces.append(VoronoiFace(structure.get_atom(0),
                                             AtomImage(structure.get_atom(0),
                                                       [0, -1, 0]),
                                             radical=True))
        neighboring_faces.append(VoronoiFace(structure.get_atom(0),
                                             AtomImage(structure.get_atom(0),
                                                       [0, 0, 1]),
                                             radical=True))
        neighboring_faces.append(VoronoiFace(structure.get_atom(0),
                                             AtomImage(structure.get_atom(0),
                                                       [0, 0, -1]),
                                             radical=True))

        # Test whether they are all there.
        for f in neighboring_faces:
            direct_faces.remove(f)
        self.assertTrue(len(direct_faces) == 0)

    def test_FCC(self):
        # Create cell.
        structure = Cell()
        structure.add_atom(Atom([0, 0, 0], 0))
        structure.add_atom(Atom([0.5, 0.5, 0], 0))
        structure.add_atom(Atom([0.5, 0, 0.5], 0))
        structure.add_atom(Atom([0, 0.5, 0.5], 0))

        # Create cell for atom1.
        cell = VoronoiCell(structure.get_atom(0), radical=True)

        # Compute faces.
        images = [AtomImage(structure.get_atom(i), sc) for i in range(4) for
                  sc in VectorCombinationComputer(
                      structure.get_lattice_vectors(),
                      4.0).get_supercell_coordinates()]
        faces = cell.compute_faces(images)

        # Get direct neighbors.
        direct_faces = cell.compute_direct_neighbors(faces)
        self.assertTrue(direct_faces[0].get_face_distance() <= direct_faces[
            -1].get_face_distance())

        # Simple tests.
        self.assertEquals(12, len(direct_faces))
        self.assertEquals(len(images) - 12 - 1, len(faces))

        # Make sure direct faces match up with expectations.
        neighboring_faces = []
        neighboring_faces.append(VoronoiFace(structure.get_atom(0),
                                             AtomImage(structure.get_atom(1),
                                                       [0, 0, 0]),
                                             radical=True))
        neighboring_faces.append(VoronoiFace(structure.get_atom(0),
                                             AtomImage(structure.get_atom(1),
                                                       [0, -1, 0]),
                                             radical=True))
        neighboring_faces.append(VoronoiFace(structure.get_atom(0),
                                             AtomImage(structure.get_atom(1),
                                                       [-1, 0, 0]),
                                             radical=True))
        neighboring_faces.append(VoronoiFace(structure.get_atom(0),
                                             AtomImage(structure.get_atom(1),
                                                       [-1, -1, 0]),
                                             radical=True))
        neighboring_faces.append(VoronoiFace(structure.get_atom(0),
                                             AtomImage(structure.get_atom(2),
                                                       [-1, 0, 0]),
                                             radical=True))
        neighboring_faces.append(VoronoiFace(structure.get_atom(0),
                                             AtomImage(structure.get_atom(2),
                                                       [-1, 0, -1]),
                                             radical=True))
        neighboring_faces.append(VoronoiFace(structure.get_atom(0),
                                             AtomImage(structure.get_atom(2),
                                                       [0, 0, 0]),
                                             radical=True))
        neighboring_faces.append(VoronoiFace(structure.get_atom(0),
                                             AtomImage(structure.get_atom(2),
                                                       [0, 0, -1]),
                                             radical=True))
        neighboring_faces.append(VoronoiFace(structure.get_atom(0),
                                             AtomImage(structure.get_atom(3),
                                                       [0, 0, 0]),
                                             radical=True))
        neighboring_faces.append(VoronoiFace(structure.get_atom(0),
                                             AtomImage(structure.get_atom(3),
                                                       [0, 0, -1]),
                                             radical=True))
        neighboring_faces.append(VoronoiFace(structure.get_atom(0),
                                             AtomImage(structure.get_atom(3),
                                                       [0, -1, 0]),
                                             radical=True))
        neighboring_faces.append(VoronoiFace(structure.get_atom(0),
                                             AtomImage(structure.get_atom(3),
                                                       [0, -1, -1]),
                                             radical=True))
        # Test whether they are all there.
        for f in neighboring_faces:
            direct_faces.remove(f)
        self.assertTrue(len(direct_faces) == 0)

    def test_intersection(self):
        # Create cell.
        structure = Cell()
        structure.add_atom(Atom([0, 0, 0], 0))

        # Create cell for atom1.
        cell = VoronoiCell(structure.get_atom(0), radical=True)

        # Make sure direct faces match up with expectations.
        neighboring_faces = []
        neighboring_faces.append(AtomImage(structure.get_atom(0), [1, 0, 0]))
        neighboring_faces.append(AtomImage(structure.get_atom(0), [-1, 0, 0]))
        neighboring_faces.append(AtomImage(structure.get_atom(0), [0, 1, 0]))
        neighboring_faces.append(AtomImage(structure.get_atom(0), [0, -1, 0]))
        neighboring_faces.append(AtomImage(structure.get_atom(0), [0, 0, 2]))
        neighboring_faces.append(AtomImage(structure.get_atom(0), [0, 0, -1]))

        # Assemble cell.
        cell.compute_cell_helper(neighboring_faces)

        # Perform cut.
        cut_face = VoronoiFace(cell.get_atom(), AtomImage(cell.get_atom(),
                    [0, 0, 1]), radical=True)
        cell.compute_intersection(cut_face)

        # Check results.
        self.assertTrue(cut_face.is_closed())
        self.assertEquals(4, cut_face.n_edges())
        self.assertTrue(cell.geometry_is_valid())
        self.assertEquals(6, cell.n_faces())
        self.assertAlmostEquals(1.0, cell.get_volume(), delta=1e-6)

    def test_edge_replacement(self):
        # Create cell.
        structure = Cell()
        structure.add_atom(Atom([0, 0, 0], 0))

        # Create cell for atom1.
        cell = VoronoiCell(structure.get_atom(0), radical=True)

        # Compute faces.
        images = [AtomImage(structure.get_atom(0), sc) for sc in
                      VectorCombinationComputer(
                          structure.get_lattice_vectors(),
                          1.1).get_supercell_coordinates()]
        cell.compute_cell_helper(images)

        # Make sure it turned out OK.
        self.assertAlmostEquals(1.0, cell.get_volume(), delta=1e-6)

        # Find position of atom that will take corner off.
        p = Plane(p1=(0.4, 0.5, 0.5), p2=(0.5, 0.4, 0.5), p3=(0.5, 0.5, 0.4),
                  tolerance=1e-6)
        atm_pos = p.project([0, 0, 0])
        atm_pos *= 2
        structure.add_atom(Atom(atm_pos, 0))

        # Cut off the corner.
        cell.compute_intersection(VoronoiFace(cell.get_atom(), AtomImage(
            structure.get_atom(1), [0, 0, 0]), radical=False))
        vol = cell.get_volume()
        self.assertEquals(7, cell.n_faces())

        # Compute a cell that will cut off just slightly more area.
        p = Plane(p1=(0.4, 0.5, 0.5), p2=(0.5, 0.4, 0.5), p3=(0.5, 0.5, 0.3),
                  tolerance=1e-6)
        atm_pos = p.project([0, 0, 0])
        atm_pos *= 2
        structure.add_atom(Atom(atm_pos, 0))
        cell.compute_intersection(VoronoiFace(cell.get_atom(), AtomImage(
            structure.get_atom(2), [0, 0, 0]), radical=False))
        self.assertEquals(7, cell.n_faces())
        self.assertTrue(cell.geometry_is_valid())
        self.assertTrue(cell.get_volume() < vol)

    def test_vertex_replacement(self):
        # Create cell.
        structure = Cell()
        structure.add_atom(Atom([0, 0, 0], 0))

        # Create cell for atom1.
        cell = VoronoiCell(structure.get_atom(0), radical=True)

        # Compute faces.
        images = [AtomImage(structure.get_atom(0), sc) for sc in
                  VectorCombinationComputer(
                      structure.get_lattice_vectors(),
                      1.1).get_supercell_coordinates()]
        cell.compute_cell_helper(images)

        # Make sure it turned out OK.
        self.assertAlmostEquals(1.0, cell.get_volume(), delta=1e-6)

        # Find position of atom that will take corner off.
        p = Plane(p1=(0.4, 0.5, 0.5), p2=(0.5, 0.4, 0.5), p3=(0.5, 0.5, 0.4),
                  tolerance=1e-6)
        atm_pos = p.project([0, 0, 0])
        atm_pos *= 2
        structure.add_atom(Atom(atm_pos, 0))

        # Cut off the corner.
        cell.compute_intersection(VoronoiFace(cell.get_atom(), AtomImage(
            structure.get_atom(1), [0, 0, 0]), radical=False))
        vol = cell.get_volume()
        self.assertEquals(7, cell.n_faces())

        # Compute a cell that will cut off just slightly more area.
        p = Plane(p1=(0.4, 0.5, 0.5), p2=(0.5, 0.35, 0.5), p3=(0.5, 0.5, 0.35),
                  tolerance=1e-6)
        atm_pos = p.project([0, 0, 0])
        atm_pos *= 2
        structure.add_atom(Atom(atm_pos, 0))
        new_face = VoronoiFace(cell.get_atom(), AtomImage(
            structure.get_atom(2), [0, 0, 0]), radical=False)
        self.assertTrue(cell.compute_intersection(new_face))
        self.assertEquals(7, cell.n_faces())
        self.assertTrue(cell.get_volume() < vol)
        self.assertTrue(cell.geometry_is_valid())

        # Remove that face.
        cell.remove_face(new_face)
        self.assertEquals(6, cell.n_faces())
        self.assertEquals(6, cell.get_polyhedron_shape()[4])
        self.assertAlmostEquals(1.0, cell.get_volume(), delta=1e-6)
        self.assertTrue(cell.geometry_is_valid())


"""
The cheml.visualization module includes (please click on links adjacent to function names for more information):
    - scatter2D: :func:`~cheml.visualization.scatter2D`
    - hist: :func:`~cheml.visualization.hist`
    - decorator: :func:`~cheml.visualization.decorator`
    - SavePlot: :func:`~cheml.visualization.SavePlot`

"""

from .visualization import scatter2D
from .visualization import hist
from .visualization import SavePlot
from .visualization import decorator


__all__ = [
    'scatter2D',
    'hist',
    'decorator',
    'SavePlot',
]



import warnings

import pandas as pd
import numpy as np

from ..utils import check_object_col

class MissingValues(object):
    """
    find missing values and interpolate/replace or remove them.

    Parameters
    ----------
    strategy: string, optional (default="ignore_row")
        
        list of strategies:
        - interpolate: interpolate based on sorted target values
        - zero: set to the zero
        - ignore_row: remove the entire row in data and target
        - ignore_column: remove the entire column in data and target

    string_as_null: boolean, optional (default=True)
        If True non numeric elements are considered to be null in computations.
    
    missing_values: list, optional (default=None)
        where you define specific formats of missing values. It is a list of string, float or integer values.

    inf_as_null: boolean, optional (default=True)
        If True inf and -inf elements are considered to be null in computations.

    Returns
    -------
    data frame
    mask: Only if strategy = ignore_row. Mask is a binary pandas series which stores the information regarding removed
    """
    def __init__(self, strategy="ignore_row", string_as_null = True,
                 inf_as_null = True, missing_values = None):
        self.strategy = strategy
        self.string_as_null = string_as_null
        self.inf_as_null = inf_as_null
        self.missing_values = missing_values
        
    def fit_transform(self, df):
        """
        use fit_transform for:
            - replace missing values with nan.
            - drop columns with all nan values.
            - fill nan values with the specified strategy.

        Parameters
        ----------
        df: pandas data frame

        Attributes
        ----------
        mask: binary pandas series, only if strategy = 'ignore_row' or 'ignore_column'
            mask is a binary vector whose length is the number of rows/indices in the df. The index of each bit shows
            if the row/column in the same position has been removed or not.
            The goal is keeping track of removed rows/columns to change the target data frame or other input data frames based
            on that. The mask can later be used in the transform method to change other data frames in the same way.
        """
        if self.inf_as_null == True:
            df.replace([np.inf, -np.inf,'inf','-inf'], np.nan, True)
        if self.string_as_null == True:
            df = df.convert_objects(convert_numeric=True)
        if isinstance(self.missing_values, (list, tuple)):
            for pattern in self.missing_values:
                df.replace(pattern, np.nan, True)

        df = check_object_col(df, 'df')
        # drop null columns
        df.dropna(axis=1, how='all', inplace=True)

        if self.strategy == 'zero':
            for col in df.columns:
                df[col].fillna(value=0,inplace=True)
            return df
        elif self.strategy == 'ignore_row':
            dfi = df.index
            df.dropna(axis=0, how='any', inplace=True)
            mask=[i in df.index for i in dfi]
            self.mask = pd.Series(mask, index=dfi)
            # self.mask = pd.notnull(df).all(1)
            # df = df[self.mask]
            return df
        elif self.strategy == 'ignore_column':
            dfc = df.columns
            df.dropna(axis=1, how='any', inplace=True)
            mask=[i in df.columns for i in dfc]
            self.mask = pd.Series(mask, index=dfc)
            # self.mask = pd.notnull(df).all(0)
            # df = df.T[self.mask].T
            return df
        elif self.strategy == 'interpolate':
            df = df.interpolate()
            df.fillna(method='ffill',axis=1, inplace=True) # because of nan in the first and last element of column
            return df
        else:
            msg = "Wrong strategy has been passed"
            raise TypeError(msg)

    def transform(self, df):
        """
        Only if the class is fitted with 'ignore_row' or 'ignore_column' strategies.

        Parameters
        ----------
        df: pandas dataframe

        Returns
        -------
        transformed data frame based on the mask vector from fit_transform method.
        """
        if self.strategy == 'ignore_row':
            return df[self.mask]
        elif self.strategy == 'ignore_column':
            return df.loc[:,self.mask]
        else:
            msg = "The transform method doesn't change the dataframe if strategy='zero' or 'interpolate'. You should fit_transform the new dataframe with those methods."
            warnings.warn(msg)



import unittest
from .....models.regression.crystal.PRDFRegression import PRDFRegression
from .....vassal.data.Atom import Atom
from .....vassal.data.Cell import Cell

class testPRDFRegression(unittest.TestCase):
    def test_distance(self):
        # Make two simple structures.
        structure1 = Cell()
        structure1.add_atom(Atom([0, 0, 0], 0))
        structure1.set_type_name(0, "Al")

        structure2 = Cell()
        structure2.add_atom(Atom([0, 0, 0], 0))
        structure2.set_type_name(0, "Ni")

        r = PRDFRegression()

        # Compute representations of each structure.
        rep1 = r.compute_representation(structure1)
        rep2 = r.compute_representation(structure2)

        # Check that similarity between identical structures is 1.0.
        self.assertAlmostEquals(1.0, r.compute_similarity(rep1, rep1),
                                delta=1e-6)
        self.assertAlmostEquals(1.0, r.compute_similarity(rep2, rep2),
                                delta=1e-6)

        # Check symmetry.
        self.assertAlmostEquals(r.compute_similarity(rep1, rep2),
                                r.compute_similarity(rep2, rep1),
                                delta=1e-6)

        # Check that similarity between these structures is less than 1.0.
        self.assertTrue(r.compute_similarity(rep1, rep2) < 1.0)

        # Check that the similarity is, in fact, 0
        self.assertAlmostEquals(0, r.compute_similarity(rep1, rep2),
                                delta=1e-6)